#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Property Interrogator Tool

@author: alee
"""
import pytest
import re

from pyomo.environ import ConcreteModel, units as pyunits
from pyomo.util.check_units import assert_units_equivalent

from idaes.core import FlowsheetBlock, MaterialBalanceType, LiquidPhase, Solute
from idaes.models.unit_models import Flash, HeatExchanger1D
from idaes.models.unit_models.pressure_changer import (
    PressureChanger,
    ThermodynamicAssumption,
)
from idaes.models.unit_models.separator import Separator, SplittingType
from idaes.models.properties.interrogator import PropertyInterrogatorBlock
from idaes.core.util.exceptions import ConfigurationError


@pytest.mark.unit
def test_interrogator_parameter_block():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    # Check that parameter block has expected attributes
    assert isinstance(m.fs.params.required_properties, dict)
    assert len(m.fs.params.required_properties) == 0


@pytest.mark.unit
def test_units_metadata():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    units_meta = m.fs.params.get_metadata().get_derived_units

    assert_units_equivalent(units_meta("length"), pyunits.m)
    assert_units_equivalent(units_meta("pressure"), pyunits.Pa)


@pytest.mark.unit
def test_interrogator_state_block_methods():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    # Check get_term methods return an unindexed dummy var
    assert m.fs.props[0].get_material_flow_terms("Liq", "A") is m.fs.props[0]._dummy_var
    assert m.fs.props[0].get_enthalpy_flow_terms("Liq") is m.fs.props[0]._dummy_var
    assert (
        m.fs.props[0].get_material_density_terms("Liq", "A") is m.fs.props[0]._dummy_var
    )
    assert m.fs.props[0].get_energy_density_terms("Liq") is m.fs.props[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.params.required_properties == {
        "material flow terms": ["fs.props"],
        "material density terms": ["fs.props"],
        "enthalpy flow terms": ["fs.props"],
        "energy density terms": ["fs.props"],
    }


@pytest.mark.unit
def test_interrogator_state_block_unindexed_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    # Check get_term methods return an unindexed dummy var
    assert m.fs.props[0].prop_unindexed is m.fs.props[0]._dummy_var

    # Call again to make sure duplicates are skipped in required_properties
    assert m.fs.props[0].prop_unindexed is m.fs.props[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.params.required_properties == {"prop_unindexed": ["fs.props"]}


@pytest.mark.unit
def test_interrogator_state_block_phase_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    # Check get_term methods return an unindexed dummy var
    assert m.fs.props[0].prop_phase["Liq"] is m.fs.props[0]._dummy_var_phase["Liq"]
    assert m.fs.props[0].prop_phase["Vap"] is m.fs.props[0]._dummy_var_phase["Vap"]

    # Check that get_term calls were logged correctly
    assert m.fs.params.required_properties == {"prop_phase": ["fs.props"]}


@pytest.mark.unit
def test_interrogator_state_block_comp_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    # Check get_term methods return an unindexed dummy var
    assert m.fs.props[0].prop_comp["A"] is m.fs.props[0]._dummy_var_comp["A"]
    assert m.fs.props[0].prop_comp["B"] is m.fs.props[0]._dummy_var_comp["B"]

    # Check that get_term calls were logged correctly
    assert m.fs.params.required_properties == {"prop_comp": ["fs.props"]}


@pytest.mark.unit
def test_interrogator_state_block_phase_comp_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    # Check get_term methods return an unindexed dummy var
    assert (
        m.fs.props[0].prop_phase_comp["Liq", "A"]
        is m.fs.props[0]._dummy_var_phase_comp["Liq", "A"]
    )
    assert (
        m.fs.props[0].prop_phase_comp["Vap", "B"]
        is m.fs.props[0]._dummy_var_phase_comp["Vap", "B"]
    )

    # Check that get_term calls were logged correctly
    assert m.fs.params.required_properties == {"prop_phase_comp": ["fs.props"]}


@pytest.mark.unit
def test_interrogator_report_method():
    # Display method should return an TypeError
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    with pytest.raises(
        TypeError,
        match="Models constructed using the Property "
        "Interrogator package should not be used for report "
        "methods.",
    ):
        m.fs.props.report()


@pytest.mark.unit
def test_interrogator_initialize_method():
    # Initialize method should return an TypeError
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.props = m.fs.params.build_state_block([0])

    with pytest.raises(
        TypeError,
        match="Models constructed using the Property "
        "Interrogator package cannot be used to solve a "
        "flowsheet. Please rebuild your flowsheet using a "
        "valid property package.",
    ):
        m.fs.props.initialize()


@pytest.fixture(scope="module")
def model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True, time_units=pyunits.s)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.P01 = PressureChanger(
        property_package=m.fs.params,
        thermodynamic_assumption=ThermodynamicAssumption.isentropic,
    )

    m.fs.HX02 = HeatExchanger1D(
        hot_side_name="shell_side",
        cold_side_name="tube_side",
        shell_side={"property_package": m.fs.params},
        tube_side={"property_package": m.fs.params},
    )

    m.fs.F03 = Flash(property_package=m.fs.params)

    return m


@pytest.mark.component
def test_interrogate_flowsheet(model):
    assert model.fs.params.required_properties == {
        "material flow terms": ["fs.P01", "fs.HX02", "fs.F03"],
        "enthalpy flow terms": ["fs.P01", "fs.HX02", "fs.F03"],
        "material density terms": ["fs.P01", "fs.HX02"],
        "energy density terms": ["fs.P01", "fs.HX02"],
        "pressure": ["fs.P01", "fs.HX02", "fs.F03"],
        "temperature": ["fs.HX02"],
        "entr_mol": ["fs.P01"],
    }


@pytest.mark.component
def test_list_required_properties(model):
    prop_list = model.fs.params.list_required_properties()

    assert prop_list == [
        "material density terms",
        "material flow terms",
        "enthalpy flow terms",
        "energy density terms",
        "pressure",
        "entr_mol",
        "temperature",
    ]


@pytest.mark.unit
def test_list_models_requiring_property(model):
    for k in model.fs.params.required_properties.keys():
        model_list = model.fs.params.list_models_requiring_property(k)

        if k == "entr_mol":
            assert model_list == ["fs.P01"]
        elif k == "temperature":
            assert model_list == ["fs.HX02"]
        elif k in ["material density terms", "energy density terms"]:
            assert model_list == ["fs.P01", "fs.HX02"]
        else:
            assert model_list == ["fs.P01", "fs.HX02", "fs.F03"]


@pytest.mark.unit
def test_list_properties_required_by_model_by_name(model):
    prop_list = model.fs.params.list_properties_required_by_model("fs.P01")
    assert prop_list == [
        "material density terms",
        "material flow terms",
        "enthalpy flow terms",
        "energy density terms",
        "pressure",
        "entr_mol",
    ]


@pytest.mark.unit
def test_list_properties_required_by_model_by_object(model):
    prop_list = model.fs.params.list_properties_required_by_model(model.fs.P01)
    assert prop_list == [
        "material density terms",
        "material flow terms",
        "enthalpy flow terms",
        "energy density terms",
        "pressure",
        "entr_mol",
    ]


@pytest.mark.unit
def test_list_properties_required_by_model_invalid_model(model):
    with pytest.raises(ValueError):
        model.fs.params.list_properties_required_by_model("foo")


@pytest.mark.unit
def test_print_required_properties(model, capsys):
    model.fs.params.print_required_properties()

    captured = capsys.readouterr()
    assert (
        captured.out
        == """
==========================================================================
Property Interrogator Summary

The Flowsheet requires the following properties (times required):

    material density terms                                               2
    material flow terms                                                  3
    enthalpy flow terms                                                  3
    energy density terms                                                 2
    pressure                                                             3
    entr_mol                                                             1
    temperature                                                          1

Note: User constraints may require additional properties which are not
reported here.
"""
    )


@pytest.mark.unit
def test_print_models_requiring_property(model, capsys):
    model.fs.params.print_models_requiring_property("pressure")

    captured = capsys.readouterr()
    assert (
        captured.out
        == """
The following models in the Flowsheet require pressure:
    fs.P01
    fs.HX02
    fs.F03
"""
    )


@pytest.mark.unit
def test_print_properties_reqruied_by_model(model, capsys):
    model.fs.params.print_properties_required_by_model("fs.P01")

    captured = capsys.readouterr()
    assert (
        captured.out
        == """
The following properties are required by model fs.P01:
    material density terms
    material flow terms
    enthalpy flow terms
    energy density terms
    pressure
    entr_mol
"""
    )


# =============================================================================
# Extra checks to make sure Interrogator works with all Separator options
# All that really matters here is that these run, but do asserts anyway
@pytest.mark.unit
def test_Separator_1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.S01 = Separator(
        property_package=m.fs.params,
        material_balance_type=MaterialBalanceType.componentPhase,
        split_basis=SplittingType.totalFlow,
        outlet_list=["a", "B", "c"],
        ideal_separation=False,
        has_phase_equilibrium=False,
    )

    assert len(m.fs.params.required_properties) == 3
    for k in m.fs.params.required_properties.keys():
        assert k in ["material flow terms", "pressure", "temperature"]


@pytest.mark.unit
def test_Separator_2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.S01 = Separator(
        property_package=m.fs.params,
        material_balance_type=MaterialBalanceType.componentPhase,
        split_basis=SplittingType.phaseFlow,
        outlet_list=["a", "B", "c"],
        ideal_separation=False,
        has_phase_equilibrium=False,
    )

    assert len(m.fs.params.required_properties) == 3
    for k in m.fs.params.required_properties.keys():
        assert k in ["material flow terms", "pressure", "temperature"]


@pytest.mark.unit
def test_Separator_3():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.S01 = Separator(
        property_package=m.fs.params,
        material_balance_type=MaterialBalanceType.componentPhase,
        split_basis=SplittingType.componentFlow,
        outlet_list=["a", "B", "c"],
        ideal_separation=False,
        has_phase_equilibrium=False,
    )

    assert len(m.fs.params.required_properties) == 3
    for k in m.fs.params.required_properties.keys():
        assert k in ["material flow terms", "pressure", "temperature"]


@pytest.mark.unit
def test_ideal_Separator_1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.S01 = Separator(
        property_package=m.fs.params,
        num_outlets=4,
        ideal_separation=True,
        split_basis=SplittingType.phaseComponentFlow,
        ideal_split_map={
            ("Vap", "A"): "outlet_1",
            ("Vap", "B"): "outlet_2",
            ("Liq", "A"): "outlet_3",
            ("Liq", "B"): "outlet_4",
        },
    )

    # Ideal Separator should require no property calls
    # Only dummy state variables are required
    assert len(m.fs.params.required_properties) == 0


@pytest.mark.unit
def test_ideal_Separator_2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.S01 = Separator(
        property_package=m.fs.params,
        num_outlets=2,
        ideal_separation=True,
        split_basis=SplittingType.phaseFlow,
        ideal_split_map={"Vap": "outlet_1", "Liq": "outlet_2"},
    )

    # Ideal Separator should require no property calls
    # Only dummy state variables are required
    assert len(m.fs.params.required_properties) == 0


@pytest.mark.unit
def test_ideal_Separator_3():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()

    m.fs.S01 = Separator(
        property_package=m.fs.params,
        num_outlets=2,
        ideal_separation=True,
        split_basis=SplittingType.componentFlow,
        ideal_split_map={"A": "outlet_1", "B": "outlet_2"},
    )

    # Ideal Separator should require no property calls
    # Only dummy state variables are required
    assert len(m.fs.params.required_properties) == 0


@pytest.mark.unit
def test_interrogator_parameter_block_custom_phase_comps():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock(
        phase_list={"P1": LiquidPhase, "P2": None},
        component_list={"c1": Solute, "c2": None},
    )

    # Check that parameter block has expected attributes
    assert isinstance(m.fs.params.required_properties, dict)
    assert len(m.fs.params.required_properties) == 0
    assert m.fs.params.phase_list == ["P1", "P2"]
    assert m.fs.params.component_list == ["c1", "c2"]


@pytest.mark.unit
def test_interrogator_state_block_methods_custom_phase_comps():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock(
        phase_list={"P1": LiquidPhase, "P2": None},
        component_list={"c1": Solute, "c2": None},
    )

    m.fs.props = m.fs.params.build_state_block([0])

    # Check get_term methods return an unindexed dummy var
    assert m.fs.props[0].get_material_flow_terms("P1", "c1") is m.fs.props[0]._dummy_var
    assert m.fs.props[0].get_enthalpy_flow_terms("P1") is m.fs.props[0]._dummy_var
    assert (
        m.fs.props[0].get_material_density_terms("P1", "c1") is m.fs.props[0]._dummy_var
    )
    assert m.fs.props[0].get_energy_density_terms("P1") is m.fs.props[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.params.required_properties == {
        "material flow terms": ["fs.props"],
        "material density terms": ["fs.props"],
        "enthalpy flow terms": ["fs.props"],
        "energy density terms": ["fs.props"],
    }


@pytest.mark.unit
def test_interrogator_parameter_block_custom_phase_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    with pytest.raises(
        ConfigurationError,
        match=re.escape(
            "fs.params invalid phase type foo (for phase "
            "P1). Type must be a subclass of Phase."
        ),
    ):
        m.fs.params = PropertyInterrogatorBlock(phase_list={"P1": "foo", "P2": None})


@pytest.mark.unit
def test_interrogator_parameter_block_custom_comp_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    with pytest.raises(
        ConfigurationError,
        match=re.escape(
            "fs.params invalid component type foo (for "
            "component c1). Type must be a subclass of "
            "Component."
        ),
    ):
        m.fs.params = PropertyInterrogatorBlock(
            component_list={"c1": "foo", "c2": None}
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from copy import deepcopy
from math import sqrt

import numpy as np

from .unit_cell_lattice import UnitCell, UnitCellLattice
from ..geometry import Parallelepiped
from ..tiling import CubicTiling
from ..transform_func import ScaleFunc, RotateFunc
from ...opt import DBL_TOL
from ...util.util import ListHasPoint


class WurtziteLattice(UnitCellLattice):
    RefIAD = sqrt(3 / 8)

    # === STANDARD CONSTRUCTOR
    def __init__(self, IAD):
        RefUnitCellShape = Parallelepiped(
            np.array([1, 0, 0], dtype=float),
            np.array([0.5, sqrt(3) / 2, 0], dtype=float),
            np.array([0, 0, sqrt(8) / sqrt(3)], dtype=float),
            BotBackLeftCorner=np.array([0, 0, 0], dtype=float),
        )
        RefUnitCellTiling = CubicTiling(RefUnitCellShape)
        RefFracPositions = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, sqrt(3) / 6.0, 1.0 / sqrt(24)]),
            np.array([0.5, sqrt(3) / 6.0, 4.0 / sqrt(24)]),
            np.array([0.0, 0.0, 5.0 / sqrt(24)]),
        ]
        RefUnitCell = UnitCell(RefUnitCellTiling, RefFracPositions)
        UnitCellLattice.__init__(self, RefUnitCell)
        self._IAD = (
            WurtziteLattice.RefIAD
        )  # IAD is set correctly after calling applyTransF
        self.applyTransF(ScaleFunc(IAD / WurtziteLattice.RefIAD))
        self._NthNeighbors = [
            [
                [
                    np.array([0.0, 0.0, -3 / sqrt(24)]),
                    np.array([0.0, -sqrt(3) / 3, 1 / sqrt(24)]),
                    np.array([0.5, sqrt(3) / 6, 1 / sqrt(24)]),
                    np.array([-0.5, sqrt(3) / 6, 1 / sqrt(24)]),
                ],
                [
                    np.array([0.0, 0.0, 3 / sqrt(24)]),
                    np.array([0.0, sqrt(3) / 3, -1 / sqrt(24)]),
                    np.array([0.5, -sqrt(3) / 6, -1 / sqrt(24)]),
                    np.array([-0.5, -sqrt(3) / 6, -1 / sqrt(24)]),
                ],
                [
                    np.array([0.0, 0.0, -3 / sqrt(24)]),
                    np.array([0.0, sqrt(3) / 3, 1 / sqrt(24)]),
                    np.array([0.5, -sqrt(3) / 6, 1 / sqrt(24)]),
                    np.array([-0.5, -sqrt(3) / 6, 1 / sqrt(24)]),
                ],
                [
                    np.array([0.0, 0.0, 3 / sqrt(24)]),
                    np.array([0.0, -sqrt(3) / 3, -1 / sqrt(24)]),
                    np.array([0.5, sqrt(3) / 6, -1 / sqrt(24)]),
                    np.array([-0.5, sqrt(3) / 6, -1 / sqrt(24)]),
                ],
            ]
        ]
        self._typeDict = {0: 0, 1: 1, 4: 2, 5: 3}
        self._relativePositions = {
            0: np.array([0.0, 0.0, 0.0]),
            1: np.array([0.5, sqrt(3) / 6.0, 1.0 / sqrt(24)]),
            4: np.array([0.5, sqrt(3) / 6.0, 4.0 / sqrt(24)]),
            5: np.array([0.0, 0.0, 5.0 / sqrt(24)]),
        }

    @classmethod
    def alignedWith(cls, IAD, MI):
        if MI == "0001":
            return cls(IAD)
        elif MI == "1100":
            result = cls(IAD)
            axis = np.array([1, 0, 0], dtype=float)
            angle = np.pi / 2
            result.applyTransF(RotateFunc.fromAxisAngle(axis, angle))
            return result
        elif MI == "1120":
            result = cls(IAD)
            axis = np.array([0, 1, 0], dtype=float)
            angle = np.pi / 2
            result.applyTransF(RotateFunc.fromAxisAngle(axis, angle))
            return result
        else:
            raise ValueError(
                "WurtziteLattice.alignedWith: Input direction is not supported."
            )

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._IAD *= TransF.Scale[0]
            else:
                raise ValueError(
                    "WurtziteLattice.applyTransF: Can only scale isometrically"
                )
        UnitCellLattice.applyTransF(self, TransF)

    # === AUXILIARY METHODS
    def _getPointType(self, P):
        return int(round(P[2] * sqrt(24))) % 8

    # === PROPERTY EVALUATION METHODS
    # NOTE: inherited from UnitCellLattice
    # def isOnLattice(self,P):

    def areNeighbors(self, P1, P2):
        return np.linalg.norm(P2 - P1) <= self.IAD + DBL_TOL

    def getNeighbors(self, P, layer=1):
        RefP = self._getConvertToReference(P)
        PType = self._getPointType(RefP)
        if PType not in self._typeDict.keys():
            raise ValueError("WurtziteLattice.getNeighbors Should never reach here!")
        if layer > len(self._NthNeighbors):
            self._calculateNeighbors(layer)
        NBs = deepcopy(self._NthNeighbors[layer - 1][self._typeDict[PType]])
        for NeighP in NBs:
            NeighP += RefP
            self._convertFromReference(NeighP)
        return NBs

    def _calculateNeighbors(self, layer):
        NList = []
        for k, v in self._typeDict.items():
            tmp = [np.array([0, 0, 0], dtype=float)]
            for nb in self._NthNeighbors:
                tmp.extend(nb[v])
            NList.append(tmp)
        for _ in range(layer - len(self._NthNeighbors)):
            tmp = [[] for _ in self._typeDict.keys()]
            for k, v in self._typeDict.items():
                for P in self._NthNeighbors[len(self._NthNeighbors) - 1][v]:
                    PType = self._getPointType(P + self._relativePositions[k])
                    for Q in self._NthNeighbors[0][self._typeDict[PType]]:
                        N = P + Q
                        if not ListHasPoint(
                            NList[v], N, 0.001 * WurtziteLattice.RefIAD
                        ):
                            tmp[v].append(N)
                            NList[v].append(N)
            self._NthNeighbors.append(tmp)

    def isASite(self, P):
        RefP = self._getConvertToReference(P)
        PType = self._getPointType(RefP)
        return PType == 0 or PType == 4

    def isBSite(self, P):
        RefP = self._getConvertToReference(P)
        PType = self._getPointType(RefP)
        return PType == 1 or PType == 5

    def setDesign(self, D, AType, BType):
        for i, P in enumerate(D.Canvas.Points):
            if self.isASite(P):
                D.setContent(i, AType)
            elif self.isBSite(P):
                D.setContent(i, BType)
            else:
                raise ValueError("setDesign can not set site not on lattice")

    # === BASIC QUERY METHODS
    @property
    def IAD(self):
        return self._IAD

    def getLayerSpacing(self, MI):
        if MI == "0001":
            return self._IAD * 4 / 3
        elif MI == "1100":
            return self._IAD * sqrt(2)
        elif MI == "1120":
            return self._IAD * sqrt(8 / 3)
        else:
            raise NotImplementedError(
                "WurtziteLattice.getLayerSpacing: Input direction is not supported."
            )

    def getShellSpacing(self, MI):
        if MI == "0001":
            return self._IAD * sqrt(2)
        elif MI == "1100":
            return self._IAD * sqrt(8 / 3)
        elif MI == "1120":
            return self._IAD * sqrt(8 / 3)
        else:
            raise NotImplementedError(
                "WurtziteLattice.getShellSpacing: Input direction is not supported."
            )

    def getUniqueLayerCount(self, MI):
        if MI == "0001":
            return 2
        elif MI == "1100":
            return 2
        elif MI == "1120":
            return 1
        else:
            raise NotImplementedError(
                "WurtziteLattice.getUniqueLayerCount: Input direction is not supported."
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Flowsheet example of the 2-region bubbling fluidized bed (BFB) model
for a methane combustion with iron-oxide case study

Created: 05/04/2020

Author: Chinedu Okoli
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import python libraries
import time

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, value

# Import IDAES core modules
from idaes.core import FlowsheetBlock
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

# Import IDAES logger
import idaes.logger as idaeslog

# Import BFB unit model
from idaes.models_extra.gas_solid_contactors.unit_models.bubbling_fluidized_bed import (
    BubblingFluidizedBed,
)

# Import property packages
from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.hetero_reactions import (
    HeteroReactionParameterBlock,
)


# -----------------------------------------------------------------------------
def main():

    # ---------------------------------------------------------------------
    # Build model

    # Create a concrete model
    m = ConcreteModel()

    # Create a steady-state flowsheet
    m.fs = FlowsheetBlock(dynamic=False)

    # Set up thermo-physical and reaction properties
    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.solid_properties = SolidPhaseParameterBlock()

    m.fs.hetero_reactions = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_properties,
        gas_property_package=m.fs.gas_properties,
    )

    # Build the BFB in the flowsheet
    m.fs.BFB = BubblingFluidizedBed(
        flow_type="co_current",
        finite_elements=5,
        transformation_method="dae.collocation",
        gas_phase_config={"property_package": m.fs.gas_properties},
        solid_phase_config={
            "property_package": m.fs.solid_properties,
            "reaction_package": m.fs.hetero_reactions,
        },
    )

    # ---------------------------------------------------------------------
    # Set design and operating variables of the BFB model

    # Fix design variables
    m.fs.BFB.number_orifice.fix(2500)  # [-]
    m.fs.BFB.bed_diameter.fix(6.5)  # m
    m.fs.BFB.bed_height.fix(5)  # m

    # Fix inlet port variables for gas and solid
    m.fs.BFB.gas_inlet.flow_mol[0].fix(1567.79)  # mol/s
    m.fs.BFB.gas_inlet.temperature[0].fix(373)  # K
    m.fs.BFB.gas_inlet.pressure[0].fix(1.86e5)  # Pa = 1E5 bar
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "O2"].fix(0.2095)
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "N2"].fix(0.7808)
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.0004)
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0093)

    m.fs.BFB.solid_inlet.flow_mass[0].fix(1230.865)  # kg/s
    # Particle porosity:
    # The porosity of the OC particle at the inlet is calculated from the
    # known bulk density of the fresh OC particle (3251.75 kg/m3), and the
    # skeletal density of the fresh OC particle (calculated from the known
    # composition of the fresh particle, and the skeletal density of its
    # components [see the solids property package])
    m.fs.BFB.solid_inlet.particle_porosity[0].fix(0.27)
    m.fs.BFB.solid_inlet.temperature[0].fix(1173.9)  # K
    m.fs.BFB.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.244162011502)
    m.fs.BFB.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(0.201998299487)
    m.fs.BFB.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.553839689011)
    # ---------------------------------------------------------------------
    # Initialize reactor

    t_start = time.time()  # Run start time

    # State arguments for initializing property state blocks
    # Bubble and gas_emulsion temperatures are initialized at solid
    # temperature because thermal mass of solid >> thermal mass of gas
    blk = m.fs.BFB
    gas_phase_state_args = {
        "flow_mol": blk.gas_inlet.flow_mol[0].value,
        "temperature": blk.solid_inlet.temperature[0].value,
        "pressure": blk.gas_inlet.pressure[0].value,
        "mole_frac": {
            "O2": blk.gas_inlet.mole_frac_comp[0, "O2"].value,
            "N2": blk.gas_inlet.mole_frac_comp[0, "N2"].value,
            "CO2": blk.gas_inlet.mole_frac_comp[0, "CO2"].value,
            "H2O": blk.gas_inlet.mole_frac_comp[0, "H2O"].value,
        },
    }
    solid_phase_state_args = {
        "flow_mass": blk.solid_inlet.flow_mass[0].value,
        "particle_porosity": blk.solid_inlet.particle_porosity[0].value,
        "temperature": blk.solid_inlet.temperature[0].value,
        "mass_frac": {
            "Fe2O3": blk.solid_inlet.mass_frac_comp[0, "Fe2O3"].value,
            "Fe3O4": blk.solid_inlet.mass_frac_comp[0, "Fe3O4"].value,
            "Al2O3": blk.solid_inlet.mass_frac_comp[0, "Al2O3"].value,
        },
    }

    print()
    print("Apply scaling transformation")
    # Scale the model by applying scaling transformation
    # This reduces ill conditioning of the model
    iscale.calculate_scaling_factors(m)

    print()
    print("Initialize the model")
    m.fs.BFB.initialize(
        outlvl=idaeslog.INFO,
        gas_phase_state_args=gas_phase_state_args,
        solid_phase_state_args=solid_phase_state_args,
    )

    t_initialize = time.time()  # Initialization time

    # ---------------------------------------------------------------------
    # Final solve

    # Create a solver
    solver = get_solver()

    solver.solve(m.fs.BFB, tee=True)

    t_simulation = time.time()  # Simulation time

    print("\n")
    print("----------------------------------------------------------")
    print("Total initialization time: ", value(t_initialize - t_start), " s")
    print("----------------------------------------------------------")

    print("\n")
    print("----------------------------------------------------------")
    print("Total simulation time: ", value(t_simulation - t_start), " s")
    print("----------------------------------------------------------")

    return m


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    m = main()

    # Plot some results
    print()
    # TODO: Fix this
    # pylint: disable-next=protected-access
    stream_table = m.fs.BFB._get_stream_table_contents()
    print(stream_table)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains miscellaneous utility functions for use in IDAES models.
"""

import pytest

from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    Expression,
    Objective,
    Set,
    Var,
    TransformationFactory,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.common.collections import ComponentSet

from idaes.core.util.model_statistics import *
from idaes.core.util.model_statistics import _iter_indexed_block_data_objects


@pytest.mark.unit
def test_iter_indexed_block_data_objects_scalar_block():
    m = ConcreteModel()

    m.s = Set(initialize=["a", "b"])
    m.b = Block()

    m.b.v = Var(m.s, initialize=1, bounds=(0, 10))

    chk_lst = list(_iter_indexed_block_data_objects(m.b, Var, True, True))
    assert len(chk_lst) == 2
    for v in chk_lst:
        assert v.name in ["b.v[a]", "b.v[b]"]


@pytest.mark.unit
def test_iter_indexed_block_data_objects_indexed_block():
    m = ConcreteModel()

    m.s = Set(initialize=["a", "b"])
    m.b = Block(m.s)

    m.b["a"].v = Var(m.s, initialize=1, bounds=(0, 10))
    m.b["b"].v = Var(m.s, initialize=1, bounds=(0, 10))

    chk_lst = list(_iter_indexed_block_data_objects(m.b, Var, True, True))
    assert len(chk_lst) == 4
    for v in chk_lst:
        assert v.name in ["b[a].v[a]", "b[a].v[b]", "b[b].v[a]", "b[b].v[b]"]


# Author: Andrew Lee
@pytest.fixture()
def m():
    m = ConcreteModel()

    m.s = Set(initialize=["a", "b"])
    m.cs = ContinuousSet(bounds=(0, 1))

    m.v = Var(m.cs, initialize=1, bounds=(0, 10))
    m.dv = DerivativeVar(m.v)

    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m, nfe=10, wrt=m.cs, scheme="BACKWARD")

    m.e = Expression(expr=m.v[1])

    m.b1 = Block()
    m.b1.v1 = Var(initialize=1)
    m.b1.v2 = Var(m.s, initialize=1)
    m.b1.v1.fix(1)
    m.b1.c1 = Constraint(expr=1 == m.b1.v1)
    m.b1.c2 = Constraint(expr=1 <= m.b1.v1)

    m.b1.sb = Block()
    m.b1.sb.v1 = Var(initialize=1)
    m.b1.sb.v2 = Var(m.s, initialize=1)
    m.b1.sb.v1.fix(1)
    m.b1.sb.e1 = Expression(expr=m.b1.sb.v1)
    m.b1.sb.o1 = Objective(expr=m.b1.sb.v1)
    m.b1.sb.o2 = Objective(expr=m.b1.sb.v1)
    m.b1.sb.o2.deactivate()
    m.b1.sb.c1 = Constraint(expr=1 == m.b1.sb.v1)
    m.b1.sb.c2 = Constraint(expr=1 >= m.b1.sb.v1)

    m.b1.deactivate()

    m.b2 = Block(m.s)
    for i in m.s:
        m.b2[i].v1 = Var(initialize=1, bounds=(1, 10))
        m.b2[i].v2 = Var(m.s, initialize=1, bounds=(0, 1))
        m.b2[i].v1.fix(1)
        m.b2[i].e1 = Expression(expr=m.b2[i].v1)
        m.b2[i].c1 = Constraint(expr=2 == m.b2[i].v1)
        m.b2[i].c2 = Constraint(expr=2 <= m.b2[i].v1)

        if i == "a":
            m.b2[i].o1 = Objective(expr=m.b2[i].v1)
            m.b2[i].o2 = Objective(expr=m.b2[i].v1)
            m.b2[i].o2.deactivate()
            m.b2[i].c1.deactivate()
            m.b2[i].c2.deactivate()

    return m


# -------------------------------------------------------------------------
# Block methods
@pytest.mark.unit
def test_total_blocks_set(m):
    assert len(total_blocks_set(m)) == 5
    assert len(total_blocks_set(m.b2)) == 1


@pytest.mark.unit
def test_number_total_blocks(m):
    assert number_total_blocks(m) == 5
    assert number_total_blocks(m.b2) == 1


@pytest.mark.unit
def test_activated_blocks_set(m):
    assert len(activated_blocks_set(m)) == 3
    assert len(activated_blocks_set(m.b2)) == 1


@pytest.mark.unit
def test_number_activated_blocks(m):
    assert number_activated_blocks(m) == 3
    assert number_activated_blocks(m.b2) == 1


@pytest.mark.unit
def test_deactivated_blocks_set(m):
    assert len(deactivated_blocks_set(m)) == 2
    assert len(deactivated_blocks_set(m.b2)) == 0


@pytest.mark.unit
def test_number_deactivated_blocks(m):
    assert number_deactivated_blocks(m) == 2
    assert number_deactivated_blocks(m.b2) == 0


# -------------------------------------------------------------------------
# Basic Constraint methods
@pytest.mark.unit
def test_total_constraints_set(m):
    assert len(total_constraints_set(m)) == 14
    assert len(total_constraints_set(m.b2)) == 4


@pytest.mark.unit
def test_number_total_constraints(m):
    assert number_total_constraints(m) == 14
    assert number_total_constraints(m.b2) == 4


@pytest.mark.unit
def test_activated_constraints_set(m):
    assert len(activated_constraints_set(m)) == 12
    assert len(activated_constraints_set(m.b2)) == 2


@pytest.mark.unit
def test_number_activated_constraints(m):
    assert number_activated_constraints(m) == 12
    assert number_activated_constraints(m.b2) == 2


@pytest.mark.unit
def test_deactivated_constraints_set(m):
    assert len(deactivated_constraints_set(m)) == 2
    assert len(deactivated_constraints_set(m.b2)) == 2


@pytest.mark.unit
def test_number_deactivated_constraints(m):
    assert number_deactivated_constraints(m) == 2
    assert number_deactivated_constraints(m.b2) == 2


# -------------------------------------------------------------------------
# Equality Constraints
@pytest.mark.unit
def test_total_equalities_set(m):
    assert len(total_equalities_set(m)) == 12
    assert len(total_equalities_set(m.b2)) == 2


@pytest.mark.unit
def test_number_total_equalities(m):
    assert number_total_equalities(m) == 12
    assert number_total_equalities(m.b2) == 2


@pytest.mark.unit
def test_activated_equalities_set(m):
    assert len(activated_equalities_set(m)) == 11
    assert len(activated_equalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_activated_equalities(m):
    assert number_activated_equalities(m) == 11
    assert number_activated_equalities(m.b2) == 1


@pytest.mark.unit
def test_deactivated_equalities_set(m):
    assert len(deactivated_equalities_set(m)) == 1
    assert len(deactivated_equalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_deactivated_equalities(m):
    assert number_deactivated_equalities(m) == 1
    assert number_deactivated_equalities(m.b2) == 1


# -------------------------------------------------------------------------
# Inequality Constraints
@pytest.mark.unit
def test_total_inequalities_set(m):
    assert len(total_inequalities_set(m)) == 2
    assert len(total_inequalities_set(m.b2)) == 2


@pytest.mark.unit
def test_number_total_inequalities(m):
    assert number_total_inequalities(m) == 2
    assert number_total_inequalities(m.b2) == 2


@pytest.mark.unit
def test_activated_inequalities_set(m):
    assert len(activated_inequalities_set(m)) == 1
    assert len(activated_inequalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_activated_inequalities(m):
    assert number_activated_inequalities(m) == 1
    assert number_activated_inequalities(m.b2) == 1


@pytest.mark.unit
def test_deactivated_inequalities_set(m):
    assert len(deactivated_inequalities_set(m)) == 1
    assert len(deactivated_inequalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_deactivated_inequalities(m):
    assert number_deactivated_inequalities(m) == 1
    assert number_deactivated_inequalities(m.b2) == 1


# -------------------------------------------------------------------------
# Basic Variable Methods
# Always use ComponentSets for Vars to avoid duplication of References
# i.e. number methods should always use the ComponentSet, not a generator
@pytest.mark.unit
def test_variables_set(m):
    assert len(variables_set(m)) == 28
    assert len(variables_set(m.b2)) == 6


@pytest.mark.unit
def test_number_variables(m):
    assert number_variables(m) == 28
    assert number_variables(m.b2) == 6


@pytest.mark.unit
def test_fixed_variables_set(m):
    assert len(fixed_variables_set(m)) == 2
    assert len(fixed_variables_set(m.b2)) == 2


@pytest.mark.unit
def test_number_fixed_variables(m):
    assert number_fixed_variables(m) == 2
    assert number_fixed_variables(m.b2) == 2


@pytest.mark.unit
def test_unfixed_variables_set(m):
    assert len(unfixed_variables_set(m)) == 26
    assert len(unfixed_variables_set(m.b2)) == 4


@pytest.mark.unit
def test_number_unfixed_variables(m):
    assert number_unfixed_variables(m) == 26
    assert number_unfixed_variables(m.b2) == 4


@pytest.mark.unit
def test_variables_near_bounds_tol_deprecation(m, caplog):
    variables_near_bounds_set(m, tol=1e-3)

    msg = (
        "DEPRECATED: variables_near_bounds_generator has deprecated the tol argument. "
        "Please set abs_tol and rel_tol arguments instead.  (deprecated in "
        "2.2.0, will be removed in (or after) 3.0.0)"
    )
    assert msg.replace(" ", "") in caplog.records[0].message.replace("\n", "").replace(
        " ", ""
    )


@pytest.mark.unit
def test_variables_near_bounds_relative_deprecation(m, caplog):
    variables_near_bounds_set(m, relative=False)

    msg = (
        "DEPRECATED: variables_near_bounds_generator has deprecated the relative argument. "
        "Please set abs_tol and rel_tol arguments instead.  (deprecated in "
        "2.2.0, will be removed in (or after) 3.0.0)"
    )
    assert msg.replace(" ", "") in caplog.records[0].message.replace("\n", "").replace(
        " ", ""
    )


@pytest.mark.unit
def test_variables_near_bounds_set():
    m = ConcreteModel()
    m.v = Var(initialize=0.5, bounds=(0, 1))

    # Small value, both bounds
    # Away from bounds
    vset = variables_near_bounds_set(m)
    assert len(vset) == 0

    # Near lower bound, relative
    m.v.set_value(1e-6)
    vset = variables_near_bounds_set(m, abs_tol=1e-8, rel_tol=1e-4)
    assert len(vset) == 1

    # Near lower bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1e-4, rel_tol=1e-8)
    assert len(vset) == 1

    # Near upper bound, relative
    m.v.set_value(1 - 1e-6)
    vset = variables_near_bounds_set(m, abs_tol=1e-8, rel_tol=1e-4)
    assert len(vset) == 1

    # Near upper bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1e-4, rel_tol=1e-8)
    assert len(vset) == 1

    # Small value, lower bound
    # Away from bounds
    m.v.set_value(0.5)
    m.v.setub(None)
    vset = variables_near_bounds_set(m)
    assert len(vset) == 0

    # Near lower bound, relative
    m.v.set_value(1e-6)
    vset = variables_near_bounds_set(m, abs_tol=1e-8, rel_tol=1e-4)
    # Lower bound of 0 means relative tolerance is 0
    assert len(vset) == 0

    # Near lower bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1e-4, rel_tol=1e-8)
    assert len(vset) == 1

    # Small value, upper bound
    # Away from bounds
    m.v.set_value(0.5)
    m.v.setub(1)
    m.v.setlb(None)
    vset = variables_near_bounds_set(m)
    assert len(vset) == 0

    # Near upper bound, relative
    m.v.set_value(1 - 1e-6)
    vset = variables_near_bounds_set(m, abs_tol=1e-8, rel_tol=1e-4)
    assert len(vset) == 1

    # Near upper bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1e-4, rel_tol=1e-8)
    assert len(vset) == 1

    # Large value, both bounds
    # Relative tolerance based on magnitude of 100
    m.v.setlb(450)
    m.v.setub(550)
    m.v.set_value(500)
    vset = variables_near_bounds_set(m)
    assert len(vset) == 0

    # Near lower bound, relative
    m.v.set_value(451)
    vset = variables_near_bounds_set(m, rel_tol=1e-2)
    assert len(vset) == 1

    # Near lower bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1)
    assert len(vset) == 1

    # Near upper bound, relative
    m.v.set_value(549)
    vset = variables_near_bounds_set(m, rel_tol=1e-2)
    assert len(vset) == 1

    # Near upper bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1)
    assert len(vset) == 1

    # Large value, lower bound
    # Relative tolerance based on magnitude of 450
    m.v.setlb(450)
    m.v.setub(None)
    m.v.set_value(500)
    vset = variables_near_bounds_set(m)
    assert len(vset) == 0

    # Near lower bound, relative
    m.v.set_value(451)
    vset = variables_near_bounds_set(m, rel_tol=1e-2)
    assert len(vset) == 1

    # Near lower bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1)
    assert len(vset) == 1

    # Large value, upper bound
    # Relative tolerance based on magnitude of 550
    m.v.setlb(None)
    m.v.setub(550)
    m.v.set_value(500)
    vset = variables_near_bounds_set(m)
    assert len(vset) == 0

    # Near upper bound, relative
    m.v.set_value(549)
    vset = variables_near_bounds_set(m, rel_tol=1e-2)
    assert len(vset) == 1

    # Near upper bound, absolute
    vset = variables_near_bounds_set(m, abs_tol=1)
    assert len(vset) == 1


@pytest.mark.unit
def test_number_variables_near_bounds(m):
    assert number_variables_near_bounds(m) == 6
    assert number_variables_near_bounds(m.b2) == 6


# -------------------------------------------------------------------------
# Variables in Constraints
@pytest.mark.unit
def test_variables_in_activated_constraints_set(m):
    assert len(variables_in_activated_constraints_set(m)) == 22
    assert len(variables_in_activated_constraints_set(m.b2)) == 1


@pytest.mark.unit
def test_number_variables_in_activated_constraints(m):
    assert number_variables_in_activated_constraints(m) == 22
    assert number_variables_in_activated_constraints(m.b2) == 1


@pytest.mark.unit
def test_variables_not_in_activated_constraints_set(m):
    assert len(variables_not_in_activated_constraints_set(m)) == 6
    assert len(variables_not_in_activated_constraints_set(m.b2)) == 5


@pytest.mark.unit
def test_number_variables_not_in_activated_constraints(m):
    assert number_variables_not_in_activated_constraints(m) == 6
    assert number_variables_not_in_activated_constraints(m.b2) == 5


@pytest.mark.unit
def test_variables_in_activated_equalities_set(m):
    assert len(variables_in_activated_equalities_set(m)) == 22
    assert len(variables_in_activated_equalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_variables_in_activated_equalities(m):
    assert number_variables_in_activated_equalities(m) == 22
    assert number_variables_in_activated_equalities(m.b2) == 1


@pytest.mark.unit
def test_variables_in_activated_inequalities_set(m):
    assert len(variables_in_activated_inequalities_set(m)) == 1
    assert len(variables_in_activated_inequalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_variables_in_activated_inequalities(m):
    assert number_variables_in_activated_inequalities(m) == 1
    assert number_variables_in_activated_inequalities(m.b2) == 1


@pytest.mark.unit
def test_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert len(variables_only_in_inequalities(m)) == 1
    assert len(variables_only_in_inequalities(m.b2)) == 0


@pytest.mark.unit
def test_number_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert number_variables_only_in_inequalities(m) == 1
    assert number_variables_only_in_inequalities(m.b2) == 0


# -------------------------------------------------------------------------
# Fixed Variables in Constraints
@pytest.mark.unit
def test_fixed_variables_in_activated_equalities_set(m):
    assert len(fixed_variables_in_activated_equalities_set(m)) == 1
    assert len(fixed_variables_in_activated_equalities_set(m.b2)) == 1


@pytest.mark.unit
def test_number_fixed_variables_in_activated_equalities(m):
    assert number_fixed_variables_in_activated_equalities(m) == 1
    assert number_fixed_variables_in_activated_equalities(m.b2) == 1


@pytest.mark.unit
def test_fixed_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.v3.fix(1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert len(fixed_variables_only_in_inequalities(m)) == 1
    assert len(fixed_variables_only_in_inequalities(m.b2)) == 0


@pytest.mark.unit
def test_number_fixed_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.v3.fix(1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert number_fixed_variables_only_in_inequalities(m) == 1
    assert number_fixed_variables_only_in_inequalities(m.b2) == 0


# -------------------------------------------------------------------------
# Unused and un-Transformed Variables
@pytest.mark.unit
def test_unused_variables_set(m):
    assert len(unused_variables_set(m)) == 6
    assert len(unused_variables_set(m.b2)) == 5


@pytest.mark.unit
def test_number_unused_variables(m):
    assert number_unused_variables(m) == 6
    assert number_unused_variables(m.b2) == 5


@pytest.mark.unit
def test_fixed_unused_variables_set(m):
    assert len(fixed_unused_variables_set(m)) == 1
    assert len(fixed_unused_variables_set(m.b2)) == 1


@pytest.mark.unit
def test_number_fixed_unused_variables(m):
    assert number_fixed_unused_variables(m) == 1
    assert number_fixed_unused_variables(m.b2) == 1


@pytest.mark.unit
def test_derivative_variables_set():
    m = ConcreteModel()

    m.cs = ContinuousSet(bounds=(0, 1))

    m.v = Var(m.cs, initialize=1)
    m.dv = DerivativeVar(m.v)

    assert len(derivative_variables_set(m)) == 2

    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m, nfe=10, wrt=m.cs, scheme="BACKWARD")

    assert len(derivative_variables_set(m)) == 0


@pytest.mark.unit
def test_number_derivative_variables():
    m = ConcreteModel()

    m.cs = ContinuousSet(bounds=(0, 1))

    m.v = Var(m.cs, initialize=1)
    m.dv = DerivativeVar(m.v)

    assert number_derivative_variables(m) == 2

    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m, nfe=10, wrt=m.cs, scheme="BACKWARD")

    assert number_derivative_variables(m) == 0


# -------------------------------------------------------------------------
# Objective methods
@pytest.mark.unit
def test_total_objectives_set(m):
    assert len(total_objectives_set(m)) == 2
    assert len(total_objectives_set(m.b2)) == 2


@pytest.mark.unit
def test_number_total_objectives(m):
    assert number_total_objectives(m) == 2
    assert number_total_objectives(m.b2) == 2


@pytest.mark.unit
def test_activated_objectives_set(m):
    assert len(activated_objectives_set(m)) == 1


@pytest.mark.unit
def test_number_activated_objectives(m):
    assert number_activated_objectives(m) == 1
    assert number_activated_objectives(m.b2) == 1


@pytest.mark.unit
def test_deactivated_objectives_set(m):
    assert len(deactivated_objectives_set(m)) == 1
    assert len(deactivated_objectives_set(m.b2)) == 1


@pytest.mark.unit
def test_number_deactivated_objectives(m):
    assert number_deactivated_objectives(m) == 1
    assert number_deactivated_objectives(m.b2) == 1


# -------------------------------------------------------------------------
# Expression methods
@pytest.mark.unit
def test_expressions_set(m):
    assert len(expressions_set(m)) == 3
    assert len(expressions_set(m.b2)) == 2


@pytest.mark.unit
def test_number_expressions(m):
    assert number_expressions(m) == 3
    assert number_expressions(m.b2) == 2


# -------------------------------------------------------------------------
# Other model statistics
@pytest.mark.unit
def test_degrees_of_freedom(m):
    assert degrees_of_freedom(m) == 10
    assert degrees_of_freedom(m.b2) == -1


@pytest.mark.unit
def test_large_residuals_set(m):
    # Initialize derivative var values so no errors occur
    for v in m.dv.keys():
        m.dv[v] = 0
    assert len(large_residuals_set(m)) == 2


@pytest.mark.unit
def test_number_large_residuals(m):
    # Initialize derivative var values so no errors occur
    for v in m.dv.keys():
        m.dv[v] = 0
    assert number_large_residuals(m) == 2


@pytest.mark.unit
def test_active_variables_in_deactivated_blocks_set(m):
    assert len(active_variables_in_deactivated_blocks_set(m)) == 0
    assert (
        len(active_variables_in_deactivated_blocks_set(m.b2)) == 1
    )  # TODO: Not sure why?

    m.c = Constraint(expr=m.b1.v1 >= 2)

    assert len(active_variables_in_deactivated_blocks_set(m)) == 1
    assert (
        len(active_variables_in_deactivated_blocks_set(m.b2)) == 1
    )  # TODO: Not sure why?


@pytest.mark.unit
def test_number_active_variables_in_deactivated_blocks(m):
    assert number_active_variables_in_deactivated_blocks(m) == 0
    assert (
        number_active_variables_in_deactivated_blocks(m.b2) == 1
    )  # TODO: Not sure why?

    m.c = Constraint(expr=m.b1.v1 >= 2)

    assert number_active_variables_in_deactivated_blocks(m) == 1
    assert (
        number_active_variables_in_deactivated_blocks(m.b2) == 1
    )  # TODO: Not sure why?


# -------------------------------------------------------------------------
# Reporting methods
@pytest.mark.unit
def test_report_statistics(m):
    report_statistics(m)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Install IDAES example files locally.
"""
import sys

# third-party
import click

# package
from idaes.commands import cb

MESSAGE = """
The get-examples command has been removed.

The IDAES examples are now available as the `idaes-examples` package in PyPI.
To install, run the following command: 

     pip install idaes-examples 
     
This also installs the 'idaesx' command that can be used to browse the Jupyter
notebooks. The simplest way to do this is with the embedded desktop UI:

    idaesx gui
    
For more details, see the IDAES documentation at https://idaes-pse.readthedocs.io/
and the idaes-examples page on PyPI at https://pypi.org/project/idaes-examples/
"""


@cb.command(
    name="get-examples", help="(legacy) install IDAES example Jupyter notebooks"
)
def get_examples():
    """Legacy get-examples command."""
    click.echo(MESSAGE)
    sys.exit(0)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example for using the eNRTL model for aqueous solutions of NaCl and KCl.

Reference:

[1] Local Composition Model for Excess Gibbs Energy of Electrolyte Systems, Pt 1.
Chen, C.-C., Britt, H.I., Boston, J.F., Evans, L.B.,
AIChE Journal, 1982, Vol. 28(4), pgs. 588-596

[2] Song, Y. and Chen, C.-C., Symmetric Electrolyte Nonrandom Two-Liquid Activity
Coefficient Model, Ind. Eng. Chem. Res., 2009, Vol. 48, pgs. 7788–7797

[3] New Data on Activity Coefficients of Potassium, Nitrate, and Chloride Ions
in Aqueous Solutions of KNO3 and KCl by Ion Selective Electrodes
Dash, D., Kumar, S., Mallika, C., Kamachi Mudali, U.,
ISRN Chemical Engineering, 2012, doi:10.5402/2012/730154

Author: Andrew Lee
"""
import pyomo.environ as pyo

from idaes.core import AqueousPhase, Solvent, Apparent, Anion, Cation
from idaes.models.properties.modular_properties.eos.enrtl import ENRTL
from idaes.models.properties.modular_properties.eos.enrtl_reference_states import (
    Symmetric,
)
from idaes.models.properties.modular_properties.base.generic_property import (
    StateIndex,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.pure.electrolyte import (
    relative_permittivity_constant,
)


def constant_density(b, *args, **kwargs):
    """Assume constant density of pure water"""
    return 1000 / 18e-3 * pyo.units.mol / pyo.units.m**3


configuration = {
    "components": {
        "H2O": {
            "type": Solvent,
            "dens_mol_liq_comp": constant_density,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (18e-3, pyo.units.kg / pyo.units.mol),
                "relative_permittivity_liq_comp": 78.54,
            },
        },
        "NaCl": {"type": Apparent, "dissociation_species": {"Na+": 1, "Cl-": 1}},
        "KCl": {"type": Apparent, "dissociation_species": {"K+": 1, "Cl-": 1}},
        "Na+": {"type": Cation, "charge": +1},
        "K+": {"type": Cation, "charge": +1},
        "Cl-": {"type": Anion, "charge": -1},
    },
    "phases": {
        "Liq": {
            "type": AqueousPhase,
            "equation_of_state": ENRTL,
            "equation_of_state_options": {"reference_state": Symmetric},
        }
    },
    "base_units": {
        "time": pyo.units.s,
        "length": pyo.units.m,
        "mass": pyo.units.kg,
        "amount": pyo.units.mol,
        "temperature": pyo.units.K,
    },
    "state_definition": FTPx,
    "state_components": StateIndex.true,
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    "parameter_data": {
        "Liq_tau": {
            ("H2O", "Na+, Cl-"): 9.0234,
            ("Na+, Cl-", "H2O"): -4.5916,
            ("H2O", "K+, Cl-"): 8.1354,
            ("K+, Cl-", "H2O"): -4.1341,
        }
    },
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for turbine stage model.

Author: John Eslick
"""
import pytest

from pyomo.environ import ConcreteModel, units as pyunits

from idaes.core import FlowsheetBlock
from idaes.models_extra.power_generation.unit_models.helm import HelmTurbineStage
from idaes.models.properties import iapws95
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    activated_equalities_generator,
)
from idaes.core.solvers import get_solver
from idaes.models.properties.general_helmholtz import helmholtz_available

# Set up solver
solver = get_solver()


@pytest.fixture()
def build_turbine():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.turb = HelmTurbineStage(property_package=m.fs.properties)
    return m


@pytest.fixture()
def build_turbine_dyn():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.turb = HelmTurbineStage(dynamic=False, property_package=m.fs.properties)
    return m


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_basic_build(build_turbine):
    """Make a turbine model and make sure it doesn't throw exception"""
    m = build_turbine


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_initialize(build_turbine):
    """Initialize a turbine model"""
    m = build_turbine
    # set inlet
    m.fs.turb.inlet.enth_mol[0].value = 70000
    m.fs.turb.inlet.flow_mol[0].value = 15000
    m.fs.turb.inlet.pressure[0].value = 8e6
    m.fs.turb.efficiency_isentropic.fix(0.8)
    m.fs.turb.ratioP.fix(0.7)
    m.fs.turb.initialize(outlvl=4)

    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert abs(c.body() - c.lower) < 1e-4
    assert degrees_of_freedom(m) == 3  # inlet was't fixed and still shouldn't be


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_get_stream_table_contents(build_turbine):
    stable = build_turbine.fs.turb._get_stream_table_contents()

    expected = {
        "Units": {
            "Mass Flow": getattr(pyunits.pint_registry, "kg/s"),
            "Molar Flow": getattr(pyunits.pint_registry, "mol/s"),
            "Molar Enthalpy": getattr(pyunits.pint_registry, "J/mol"),
            "P": getattr(pyunits.pint_registry, "Pa"),
            "T": getattr(pyunits.pint_registry, "K"),
            "Vapor Fraction": getattr(pyunits.pint_registry, "dimensionless"),
        },
        "Inlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "Outlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
    }

    assert stable.to_dict() == expected



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example of defining a reaction package for the following reaction system:

R1)  A + B --> 2C  Power law rate reaction
R2)  B + C <-> D  Power law equilibrium reaction

Author: Andrew Lee
"""
from pyomo.environ import units as pyunits

from idaes.core import LiquidPhase, Component

from idaes.models.properties.modular_properties.state_definitions import FcTP
from idaes.models.properties.modular_properties.eos.ideal import Ideal

from idaes.models.properties.modular_properties.base.utility import ConcentrationForm
from idaes.models.properties.modular_properties.reactions.dh_rxn import constant_dh_rxn
from idaes.models.properties.modular_properties.reactions.rate_constant import arrhenius
from idaes.models.properties.modular_properties.reactions.rate_forms import (
    power_law_rate,
)
from idaes.models.properties.modular_properties.reactions.equilibrium_constant import (
    van_t_hoff,
)
from idaes.models.properties.modular_properties.reactions.equilibrium_forms import (
    power_law_equil,
)


# First, create a thermophsyical property definition that will be used
# with the reactions
import idaes.models.properties.modular_properties.pure.Perrys as Perrys

# For this example, the thermophsycial properties will only define components
# and phases, but in practice users would also need to define some properties
thermo_configuration = {
    "components": {
        "A": {
            "type": Component,
            "enth_mol_liq_comp": Perrys,
            "parameter_data": {
                "cp_mol_liq_comp_coeff": {
                    "1": (1e2, pyunits.J / pyunits.kmol / pyunits.K),
                    "2": (-2e-1, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (6.5e-4, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (0, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (0, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (5e4, pyunits.J / pyunits.mol),
            },
        },
        "B": {
            "type": Component,
            "enth_mol_liq_comp": Perrys,
            "parameter_data": {
                "cp_mol_liq_comp_coeff": {
                    "1": (1e2, pyunits.J / pyunits.kmol / pyunits.K),
                    "2": (-2e-1, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (6.5e-4, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (0, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (0, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (5e4, pyunits.J / pyunits.mol),
            },
        },
        "C": {
            "type": Component,
            "enth_mol_liq_comp": Perrys,
            "parameter_data": {
                "cp_mol_liq_comp_coeff": {
                    "1": (1e2, pyunits.J / pyunits.kmol / pyunits.K),
                    "2": (-2e-1, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (6.5e-4, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (0, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (0, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (5e4, pyunits.J / pyunits.mol),
            },
        },
        "D": {
            "type": Component,
            "enth_mol_liq_comp": Perrys,
            "parameter_data": {
                "cp_mol_liq_comp_coeff": {
                    "1": (1e2, pyunits.J / pyunits.kmol / pyunits.K),
                    "2": (-2e-1, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (6.5e-4, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (0, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (0, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (5e4, pyunits.J / pyunits.mol),
            },
        },
    },
    "phases": {"Liq": {"type": LiquidPhase, "equation_of_state": Ideal}},
    "state_definition": FcTP,
    "state_bounds": {
        "flow_mol_comp": (0, 500, 1000),
        "temperature": (273.15, 300, 450),
        "pressure": (5e4, 1e5, 1e6),
    },
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
}


# Next, create the reaction property definition which describes the system on
# reactions to be modeled.
rxn_configuration = {
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "rate_reactions": {
        "R1": {
            "stoichiometry": {("Liq", "A"): -1, ("Liq", "B"): -1, ("Liq", "C"): 2},
            "heat_of_reaction": constant_dh_rxn,
            "rate_constant": arrhenius,
            "rate_form": power_law_rate,
            "concentration_form": ConcentrationForm.moleFraction,
            "parameter_data": {
                "dh_rxn_ref": (-10000, pyunits.J / pyunits.mol),
                "arrhenius_const": (1, pyunits.mol / pyunits.m**3 / pyunits.s),
                "energy_activation": (1000, pyunits.J / pyunits.mol),
            },
        }
    },
    "equilibrium_reactions": {
        "R2": {
            "stoichiometry": {("Liq", "B"): -1, ("Liq", "C"): -1, ("Liq", "D"): 1},
            "heat_of_reaction": constant_dh_rxn,
            "equilibrium_constant": van_t_hoff,
            "equilibrium_form": power_law_equil,
            "concentration_form": ConcentrationForm.moleFraction,
            "parameter_data": {
                "dh_rxn_ref": (-20000, pyunits.J / pyunits.mol),
                "k_eq_ref": (100, None),
                "T_eq_ref": (350, pyunits.K),
            },
        }
    },
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for CLC solid phase thermo state block; tests for construction and solve
Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Var,
    Constraint,
    value,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap
from pyomo.util.subsystems import ParamSweeper
from pyomo.contrib.incidence_analysis import (
    solve_strongly_connected_components,
)

from idaes.core import FlowsheetBlock

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_large_residuals,
    fixed_variables_set,
    activated_constraints_set,
)

from idaes.core.solvers import get_solver
from idaes.core.util import scaling as iscale

from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)

# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.fixture(scope="class")
def solid_prop():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # solid properties and state inlet block
    m.fs.properties = SolidPhaseParameterBlock()

    m.fs.unit = m.fs.properties.build_state_block(
        [0], parameters=m.fs.properties, defined_state=True
    )

    m.fs.unit[0].flow_mass.fix(1)
    m.fs.unit[0].particle_porosity.fix(0.27)
    m.fs.unit[0].temperature.fix(1183.15)
    m.fs.unit[0].mass_frac_comp["Fe2O3"].fix(0.244)
    m.fs.unit[0].mass_frac_comp["Fe3O4"].fix(0.202)
    m.fs.unit[0].mass_frac_comp["Al2O3"].fix(0.554)

    return m


@pytest.mark.unit
def test_build_inlet_state_block(solid_prop):
    assert isinstance(solid_prop.fs.unit[0].dens_mass_skeletal, Var)
    assert isinstance(solid_prop.fs.unit[0].enth_mol_comp, Var)
    assert isinstance(solid_prop.fs.unit[0].enth_mass, Var)
    assert isinstance(solid_prop.fs.unit[0].cp_mol_comp, Var)
    assert isinstance(solid_prop.fs.unit[0].cp_mass, Var)


@pytest.mark.unit
def test_setInputs_state_block(solid_prop):
    assert degrees_of_freedom(solid_prop.fs.unit[0]) == 0


@pytest.fixture(scope="class")
def solid_prop_unscaled():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # solid properties and state inlet block
    m.fs.properties = SolidPhaseParameterBlock()

    m.fs.unit = m.fs.properties.build_state_block(
        [0], parameters=m.fs.properties, defined_state=True
    )

    m.fs.unit[0].flow_mass.fix(1)
    m.fs.unit[0].particle_porosity.fix(0.27)
    m.fs.unit[0].temperature.fix(1183.15)
    m.fs.unit[0].mass_frac_comp["Fe2O3"].fix(0.244)
    m.fs.unit[0].mass_frac_comp["Fe3O4"].fix(0.202)
    m.fs.unit[0].mass_frac_comp["Al2O3"].fix(0.554)

    return m


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unscaled(solid_prop_unscaled):
    orig_fixed_vars = fixed_variables_set(solid_prop_unscaled)
    orig_act_consts = activated_constraints_set(solid_prop_unscaled)

    solid_prop_unscaled.fs.unit.initialize()

    assert degrees_of_freedom(solid_prop_unscaled) == 0

    fin_fixed_vars = fixed_variables_set(solid_prop_unscaled)
    fin_act_consts = activated_constraints_set(solid_prop_unscaled)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unscaled(solid_prop_unscaled):

    assert hasattr(solid_prop_unscaled.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop_unscaled.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop_unscaled.fs.unit[0], "enth_mass")

    results = solver.solve(solid_prop_unscaled)

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.component
def test_scaling(solid_prop):
    # Calculate scaling factors

    # Construct property methods to build the constraints

    assert hasattr(solid_prop.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop.fs.unit[0], "dens_mass_particle")
    assert hasattr(solid_prop.fs.unit[0], "cp_mol_comp")
    assert hasattr(solid_prop.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mol_comp")

    # Call flow and density methods to construct flow and density expressions
    for i in solid_prop.fs.unit[0]._params.component_list:
        solid_prop.fs.unit[0].get_material_flow_terms("Sol", i)
        solid_prop.fs.unit[0].get_material_density_terms("Sol", i)
    solid_prop.fs.unit[0].get_enthalpy_flow_terms("Sol")
    solid_prop.fs.unit[0].get_energy_density_terms("Sol")

    # Calculate scaling factors now that constraints/expressions are built
    iscale.calculate_scaling_factors(solid_prop)

    # Test scaling
    assert pytest.approx(1e-2, rel=1e-5) == iscale.get_scaling_factor(
        solid_prop.fs.unit[0].dens_mass_particle
    )

    for i, c in solid_prop.fs.unit[0].material_flow_terms.items():
        assert pytest.approx(1e-2, rel=1e-5) == iscale.get_scaling_factor(c)
    for i, c in solid_prop.fs.unit[0].material_density_terms.items():
        assert pytest.approx(1e-1, rel=1e-5) == iscale.get_scaling_factor(c)
    for i, c in solid_prop.fs.unit[0].energy_density_terms.items():
        assert pytest.approx(1e-8, rel=1e-5) == iscale.get_scaling_factor(c)
    for i, c in solid_prop.fs.unit[0].enthalpy_flow_terms.items():
        assert pytest.approx(1e-9, rel=1e-5) == iscale.get_scaling_factor(c)

    assert pytest.approx(
        1e-2, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].density_particle_constraint
    )
    assert pytest.approx(
        1e-2, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].density_skeletal_constraint
    )
    for i, c in solid_prop.fs.unit[0].cp_shomate_eqn.items():
        assert pytest.approx(
            1e-6, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1e-6, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].mixture_heat_capacity_eqn
    )
    for i, c in solid_prop.fs.unit[0].enthalpy_shomate_eqn.items():
        assert pytest.approx(
            1e-6, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1e-6, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        solid_prop.fs.unit[0].mixture_enthalpy_eqn
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(solid_prop):
    orig_fixed_vars = fixed_variables_set(solid_prop)
    orig_act_consts = activated_constraints_set(solid_prop)

    solid_prop.fs.unit.initialize()

    assert degrees_of_freedom(solid_prop) == 0

    fin_fixed_vars = fixed_variables_set(solid_prop)
    fin_act_consts = activated_constraints_set(solid_prop)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve(solid_prop):

    assert hasattr(solid_prop.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mass")

    results = solver.solve(solid_prop)

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solution(solid_prop):
    assert (
        pytest.approx(3251.75, rel=1e-5)
        == solid_prop.fs.unit[0].dens_mass_skeletal.value
    )
    assert pytest.approx(1, rel=1e-5) == solid_prop.fs.unit[0].cp_mass.value
    assert pytest.approx(0.0000, rel=1e-5) == solid_prop.fs.unit[0].enth_mass.value


@pytest.mark.component
def test_units_consistent(solid_prop):

    # Construct property methods to build the constraints

    assert hasattr(solid_prop.fs.unit[0], "dens_mass_skeletal")
    assert hasattr(solid_prop.fs.unit[0], "dens_mass_particle")
    assert hasattr(solid_prop.fs.unit[0], "cp_mol_comp")
    assert hasattr(solid_prop.fs.unit[0], "cp_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mass")
    assert hasattr(solid_prop.fs.unit[0], "enth_mol_comp")

    # Call flow and density methods to construct flow and density expressions
    for i in solid_prop.fs.unit[0]._params.component_list:
        solid_prop.fs.unit[0].get_material_flow_terms("Sol", i)
        solid_prop.fs.unit[0].get_material_density_terms("Sol", i)
    solid_prop.fs.unit[0].get_enthalpy_flow_terms("Sol")
    solid_prop.fs.unit[0].get_energy_density_terms("Sol")

    assert_units_consistent(solid_prop)


@pytest.mark.unit
def test_state_vars():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = SolidPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block()

    assert isinstance(m.fs.state.flow_mass, Var)
    assert isinstance(m.fs.state.temperature, Var)
    assert isinstance(m.fs.state.particle_porosity, Var)
    assert isinstance(m.fs.state.mass_frac_comp, Var)

    assert isinstance(m.fs.state.sum_component_eqn, Constraint)

    assert len(list(m.fs.state.component_data_objects(Var))) == 6
    assert len(list(m.component_data_objects(Constraint))) == 1


@pytest.mark.unit
def test_indexed_state_block():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = SolidPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block([1, 2, 3])

    assert len([v for v in m.component_data_objects(Var) if not v.fixed]) == 18
    assert len(list(m.component_data_objects(Constraint))) == 3

    for i, state in m.fs.state.items():
        assert isinstance(state.flow_mass, Var)
        assert isinstance(state.temperature, Var)
        assert isinstance(state.particle_porosity, Var)
        assert isinstance(state.mass_frac_comp, Var)

        assert isinstance(state.sum_component_eqn, Constraint)


@pytest.mark.unit
def test_property_construction_ordered():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = SolidPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block()

    # If we construct properties in this order, variables and constraints
    # will be added one at a time.
    matching = [
        ("dens_mass_skeletal", "density_skeletal_constraint"),
        ("dens_mass_particle", "density_particle_constraint"),
        ("cp_mol_comp", "cp_shomate_eqn"),
        ("cp_mass", "mixture_heat_capacity_eqn"),
        ("enth_mol_comp", "enthalpy_shomate_eqn"),
        ("enth_mass", "mixture_enthalpy_eqn"),
    ]

    state_vars = m.fs.state.define_state_vars()
    n_state_vars = len(state_vars)
    n_vars = len(m.fs.properties._metadata.properties.list_supported_properties())
    assert len(matching) == n_vars - n_state_vars

    nvar = len(list(m.fs.state.component_data_objects(Var)))
    ncon = len(list(m.component_data_objects(Constraint)))

    for varname, conname in matching:
        assert varname not in state_vars
        var = getattr(m.fs.state, varname)
        con = getattr(m.fs.state, conname)
        dim = len(var)
        nvar += dim
        ncon += dim
        assert dim == len(con)
        assert isinstance(var, Var)
        assert isinstance(con, Constraint)
        assert len(list(m.fs.state.component_data_objects(Var))) == nvar
        assert len(list(m.component_data_objects(Constraint))) == ncon


@pytest.mark.unit
class TestProperties(unittest.TestCase):
    def _make_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = SolidPhaseParameterBlock()
        m.fs.state = m.fs.properties.build_state_block(defined_state=True)
        for var in m.fs.state.define_state_vars().values():
            var.fix()
        return m

    def test_dens_mass_skeletal(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 4
        state_values = {
            "flow_mass": [1.0 * pyunits.kg / pyunits.s] * n_scen,
            "temperature": [1200.0 * pyunits.K] * n_scen,
            "particle_porosity": [0.27] * n_scen,
            "mass_frac_comp[Fe2O3]": [1.0, 0.0, 0.0, 1.0 / 3.0],
            "mass_frac_comp[Fe3O4]": [0.0, 1.0, 0.0, 1.0 / 3.0],
            "mass_frac_comp[Al2O3]": [0.0, 0.0, 1.0, 1.0 / 3.0],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )
        kgm3 = pyunits.kg / pyunits.m**3
        target_values = {
            "dens_mass_skeletal": [
                5250.000 * kgm3,
                5000.000 * kgm3,
                3987.000 * kgm3,
                4678.061 * kgm3,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        param_sweeper = ParamSweeper(
            n_scen,
            state_values,
            output_values=target_values,
        )
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(state)

                assert number_large_residuals(state, tol=1e-8) == 0

                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                for var, val in outputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_dens_mass_particle(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 3
        state_values = {
            "flow_mass": [1.0 * pyunits.kg / pyunits.s] * n_scen,
            "temperature": [1200.0 * pyunits.K] * n_scen,
            "particle_porosity": [0.22, 0.27, 0.32],
            "mass_frac_comp[Fe2O3]": [1.0 / 3.0] * n_scen,
            "mass_frac_comp[Fe3O4]": [1.0 / 3.0] * n_scen,
            "mass_frac_comp[Al2O3]": [1.0 / 3.0] * n_scen,
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )
        kgm3 = pyunits.kg / pyunits.m**3
        target_values = {
            "dens_mass_particle": [
                3648.888 * kgm3,
                3414.985 * kgm3,
                3181.081 * kgm3,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        param_sweeper = ParamSweeper(
            n_scen,
            state_values,
            output_values=target_values,
        )
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(state)

                # Check that we have eliminated infeasibility
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs have been set properly
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                for var, val in outputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_cp(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 4
        K = pyunits.K
        state_values = {
            "flow_mass": [1.0 * pyunits.kg / pyunits.s] * n_scen,
            "temperature": [1000.0 * K, 1100 * K, 1200 * K, 1300 * K],
            "particle_porosity": [0.27] * n_scen,
            "mass_frac_comp[Fe2O3]": [1.0 / 3.0] * n_scen,
            "mass_frac_comp[Fe3O4]": [1.0 / 3.0] * n_scen,
            "mass_frac_comp[Al2O3]": [1.0 / 3.0] * n_scen,
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )
        kJmolK = pyunits.kJ / pyunits.mol / pyunits.K
        kJkgK = pyunits.kJ / pyunits.kg / pyunits.K
        target_values = {
            "cp_mol_comp[Fe2O3]": [
                0.1401 * kJmolK,
                0.1408 * kJmolK,
                0.1415 * kJmolK,
                0.1423 * kJmolK,
            ],
            "cp_mol_comp[Fe3O4]": [
                0.2008 * kJmolK,
                0.2008 * kJmolK,
                0.2008 * kJmolK,
                0.2008 * kJmolK,
            ],
            "cp_mol_comp[Al2O3]": [
                0.1249 * kJmolK,
                0.1268 * kJmolK,
                0.1285 * kJmolK,
                0.1299 * kJmolK,
            ],
            "cp_mass": [
                0.9899 * kJkgK,
                0.9975 * kJkgK,
                1.0045 * kJkgK,
                1.0108 * kJkgK,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        param_sweeper = ParamSweeper(
            n_scen,
            state_values,
            output_values=target_values,
        )
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(state)

                # Check that we have eliminated infeasibility
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs have been set properly
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                for var, val in outputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_enth(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 4
        K = pyunits.K
        state_values = {
            "flow_mass": [1.0 * pyunits.kg / pyunits.s] * n_scen,
            "temperature": [1000.0 * K, 1100 * K, 1200 * K, 1300 * K],
            "particle_porosity": [0.27] * n_scen,
            "mass_frac_comp[Fe2O3]": [1.0 / 3.0] * n_scen,
            "mass_frac_comp[Fe3O4]": [1.0 / 3.0] * n_scen,
            "mass_frac_comp[Al2O3]": [1.0 / 3.0] * n_scen,
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )
        kJmol = pyunits.kJ / pyunits.mol
        kJkg = pyunits.kJ / pyunits.kg
        target_values = {
            "enth_mol_comp[Fe2O3]": [
                101.043 * kJmol,
                115.086 * kJmol,
                129.198 * kJmol,
                143.385 * kJmol,
            ],
            "enth_mol_comp[Fe3O4]": [
                147.591 * kJmol,
                167.674 * kJmol,
                187.757 * kJmol,
                207.841 * kJmol,
            ],
            "enth_mol_comp[Al2O3]": [
                77.925 * kJmol,
                90.513 * kJmol,
                103.279 * kJmol,
                116.199 * kJmol,
            ],
            "enth_mass": [
                678.156 * kJkg,
                777.534 * kJkg,
                877.640 * kJkg,
                978.408 * kJkg,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        param_sweeper = ParamSweeper(
            n_scen,
            state_values,
            output_values=target_values,
        )
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(state)

                # Check that we have eliminated infeasibility
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs have been set properly
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                for var, val in outputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Costing models for a direct air capture (DAC) plant. The fixed-bed temperature
swing adsorption model used to represent the DAC process should be passed as an
argument to the costing functions.
"""

__author__ = "Daison Yancy Caballero"

import os
import json
import textwrap
from sys import stdout
from pandas import DataFrame

from pyomo.environ import units, Var, value
from pyomo.common.fileutils import this_file_dir
from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core import UnitModelBlock, UnitModelCostingBlock
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.tables import stream_table_dataframe_to_string

from idaes.models_extra.power_generation.costing.power_plant_capcost import (
    QGESSCosting,
    QGESSCostingData,
)

directory = this_file_dir()


def get_dac_costing(tsa, costing_case="electric_boiler"):
    if costing_case == "electric_boiler":
        _get_costing_electric_boiler(tsa)
    elif costing_case == "retrofit_ngcc":
        _get_costing_retrofit_ngcc(tsa)
    else:
        raise ConfigurationError("costing case not defined.")


def _get_costing_electric_boiler(tsa):

    # load custom costing parameters
    with open(
        os.path.join(directory, "costing_params_dac_electric_boiler.json"), "r"
    ) as f:
        costing_params = json.load(f)

    # get flowsheet
    fs = tsa.flowsheet()

    # create costing block
    fs.costing = QGESSCosting()

    # EPAT accounts for:
    # 1) Sorbent Handling, 2) Sorbent Preparation & Feed,
    # 3) Feedwater & Miscellaneous Bop Systems, 9) Cooling Water System,
    # 10) Ash/Spent Sorbent Handling System, 11) Accessory Electric Plant
    # 12) Instrumentation & Control, 13) Improvements to Site,
    # 14) Buildings & Structures, 15) DAC system

    # grouping accounts
    sorbent_makeup_accounts = [
        "1.5",
        "1.6",
        "1.7",
        "1.8",
        "1.9",
        "2.5",
        "2.6",
        "2.9",
        "10.6",
        "10.7",
        "10.9",
    ]
    raw_water_withdrawal_accounts = ["3.2", "3.4", "9.5", "14.6"]
    steam_accounts = ["3.1", "3.3", "3.5"]
    process_water_discharge_accounts = ["3.7"]
    cooling_tower_accounts = ["9.1"]
    circulating_water_accounts = ["9.2", "9.3", "9.4", "9.6", "9.7", "14.5"]
    total_auxiliary_load_accounts = [
        "11.1",
        "11.2",
        "11.3",
        "11.4",
        "11.5",
        "11.6",
        "11.7",
        "11.8",
        "11.9",
        "12.4",
        "12.5",
        "12.6",
        "12.7",
        "12.8",
        "12.9",
        "13.1",
        "13.2",
        "13.3",
        "14.4",
        "14.7",
        "14.8",
        "14.9",
        "14.10",
    ]
    dac_system_accounts = [
        "15.1",
        "15.2",
        "15.3",
        "15.4",
        "15.5",
        "15.6",
        "15.7",
        "15.8",
        "15.9",
    ]

    # reference parameters for accounts
    # sorbent makeup rate - from model
    sorbent_makeup_rate = (
        units.convert(tsa.bed_volume, to_units=units.ft**3)
        * (1 - tsa.bed_voidage)
        * tsa.number_beds
        / 0.5
        / 365
        / units.d
    )  # [ft^3/d]
    # CO2 product mass flow rate - from model
    CO2_product_mass_flow = units.convert(
        tsa.mw["CO2"] * tsa.flow_mol_co2_rich_stream[0, "CO2"],
        to_units=units.lb / units.hr,
    )  # [lb/hr]
    # raw water withdrawal flow rate - from surrogates
    raw_water_withdrawal = (
        (1.0496e-03 * CO2_product_mass_flow * units.hr / units.lb + 5.3359e01)
        * units.gal
        / units.min
    )  # [gpm]
    # process water discharge flow rate - from surrogates
    process_water_discharge = (
        (5.4007e-04 * CO2_product_mass_flow * units.hr / units.lb + 1.2000e01)
        * units.gal
        / units.min
    )  # [gpm]
    # cooling tower heat duty - from surrogates
    cooling_tower_duty = (
        (3.0797e-04 * CO2_product_mass_flow * units.hr / units.lb + 2.5000e01)
        * units.MBtu
        / units.hr
    )  # [MMBtu/hr]
    # circulating water flow_rate - from surrogates
    circulating_water_flow_rate = (
        (3.0797e-02 * CO2_product_mass_flow * units.hr / units.lb + 2.5000e03)
        * units.gal
        / units.min
    )  # [gpm]
    # mole flow rate of dac feed - from model (exhaust or air)
    flow_mol_inlet = units.convert(
        tsa.flow_mol_in_total, to_units=units.kmol / units.hr
    )  # [kmol/hr]
    # compressor auxiliary load - from surrogates
    product_compressor_auxiliary_load = (
        0.0012 * flow_mol_inlet * units.hr / units.kmol - 2.2798
    ) * units.kW  # [kW]
    # boiler auxiliary load - from surrogates
    boiler_auxiliary_load = (
        0.000335999
        * units.convert(tsa.flow_mass_steam, to_units=units.lb / units.hr)
        * units.hr
        / units.lb
        * 1e3
    ) * units.kW  # covert MW to kW
    # total auxiliary load
    # calculate with fans work + compressor for CO2 pure + boiler aux load
    total_auxiliary_load = (
        product_compressor_auxiliary_load
        + boiler_auxiliary_load
        + units.convert(tsa.compressor.unit.work_mechanical[0], to_units=units.kW)
    )
    # compressor aftercooler heat exchanger duty - from surrogates
    compressor_aftercooler_heat_duty = (
        (2e-6 * flow_mol_inlet * units.hr / units.kmol - 7e-8) * units.MBtu / units.hr
    )  # [MMBtu/hr]
    # auxiliary load for 2 beds (pressure changer from TSA model estimates
    # the power required to move air or exhaust gas to all beds)
    auxiliary_load_2_beds = (
        2
        * units.convert(tsa.compressor.unit.work_mechanical[0], to_units=units.kW)
        / tsa.number_beds
    )  # [kW]

    # sorbent makeup accounts
    fs.sorbent_makeup = UnitModelBlock()
    fs.sorbent_makeup.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": sorbent_makeup_accounts,
            "scaled_param": sorbent_makeup_rate,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # raw water withdrawal accounts
    fs.raw_water_withdrawal = UnitModelBlock()
    fs.raw_water_withdrawal.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": raw_water_withdrawal_accounts,
            "scaled_param": raw_water_withdrawal,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # steam accounts
    fs.steam = UnitModelBlock()
    fs.steam.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": steam_accounts,
            "scaled_param": units.convert(
                tsa.flow_mass_steam, to_units=units.lb / units.hr
            ),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # process water discharge accounts
    fs.process_water_discharge = UnitModelBlock()
    fs.process_water_discharge.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": process_water_discharge_accounts,
            "scaled_param": process_water_discharge,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # cooling tower accounts
    fs.cooling_tower = UnitModelBlock()
    fs.cooling_tower.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": cooling_tower_accounts,
            "scaled_param": cooling_tower_duty,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # circulating water accounts
    fs.circulating_water = UnitModelBlock()
    fs.circulating_water.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": circulating_water_accounts,
            "scaled_param": circulating_water_flow_rate,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # total plant auxiliary load accounts
    fs.total_auxiliary_load = UnitModelBlock()
    fs.total_auxiliary_load.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": total_auxiliary_load_accounts,
            "scaled_param": total_auxiliary_load,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # get EPAT costing accounts for 15) DAC system
    accounts_info = {}
    for k, v in costing_params["8"]["B"].items():
        if k in dac_system_accounts:
            accounts_info[k] = v["Account Name"].replace("DAC ", "")
    accounts = list(accounts_info.keys())

    # 15.1 - DAC adsorption/desorption vessels (cost of 120 vessels)
    vessels_account = [accounts[0]]
    fs.vessels = UnitModelBlock()
    fs.vessels.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": vessels_account,
            "scaled_param": units.convert(tsa.bed_volume, to_units=units.ft**3),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.2 - DAC CO2 Compression & Drying
    product_compression_account = [accounts[1]]
    fs.product_compression = UnitModelBlock()
    fs.product_compression.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": product_compression_account,
            "scaled_param": product_compressor_auxiliary_load,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.3 - DAC CO2 Compressor Aftercooler
    compressor_aftercooler_account = [accounts[2]]
    fs.compressor_aftercooler = UnitModelBlock()
    fs.compressor_aftercooler.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": compressor_aftercooler_account,
            "scaled_param": compressor_aftercooler_heat_duty,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.4 - DAC System Air Handling Duct and Dampers
    duct_dampers_account = [accounts[3]]
    fs.duct_dampers = UnitModelBlock()
    fs.duct_dampers.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": duct_dampers_account,
            "scaled_param": 2
            * units.convert(tsa.flow_mass_in_total_bed, to_units=units.lb / units.hr),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.5 - DAC System Air Handling Fans
    feed_fans_account = [accounts[4]]
    fs.feed_fans = UnitModelBlock()
    fs.feed_fans.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": feed_fans_account,
            "scaled_param": auxiliary_load_2_beds,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.6 - DAC Desorption Process Gas Handling System
    desorption_gas_handling_account = [accounts[5]]
    fs.desorption_gas_handling = UnitModelBlock()
    fs.desorption_gas_handling.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": desorption_gas_handling_account,
            "scaled_param": CO2_product_mass_flow,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.7 - DAC Steam Distribution System
    steam_distribution_account = [accounts[6]]
    fs.steam_distribution = UnitModelBlock()
    fs.steam_distribution.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": steam_distribution_account,
            "scaled_param": units.convert(
                tsa.flow_mass_steam, to_units=units.lb / units.hr
            ),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.8 - DAC System Controls Equipment
    controls_equipment_account = [accounts[7]]
    fs.controls_equipment = UnitModelBlock()
    fs.controls_equipment.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": controls_equipment_account,
            "scaled_param": total_auxiliary_load,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.9 - Electric Boiler
    electric_boiler_account = [accounts[8]]
    fs.electric_boiler = UnitModelBlock()
    fs.electric_boiler.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": electric_boiler_account,
            "scaled_param": units.convert(
                tsa.flow_mass_steam, to_units=units.lb / units.hr
            ),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # total plant cost
    TPC_list = {}
    for o in fs.component_objects(descend_into=True):
        # look for costing blocks
        if hasattr(o, "costing") and hasattr(o.costing, "total_plant_cost"):
            for k in o.costing.total_plant_cost.keys():
                if k not in ["15.1", "15.4", "15.5"]:
                    TPC_list[k] = o.costing.total_plant_cost[k]
                if k in ["15.1"]:
                    TPC_list[k] = o.costing.total_plant_cost[k] / 120 * tsa.number_beds
                if k in ["15.4", "15.5"]:
                    TPC_list[k] = o.costing.total_plant_cost[k] * tsa.number_beds / 2

    # Total plant cost of dac unit
    @fs.costing.Expression(doc="total TPC for TSA system in $MM")
    def total_TPC(b):
        return sum(TPC_list.values())

    # build variable costs components

    # gallon/day of water required
    water_rate = (
        units.convert(tsa.flow_mol_in_total, to_units=units.kmol / units.hr)
        * units.hr
        / units.kmol
        * 60792.407
        / 1629629
        * units.gallon
        / units.day
    )
    # ton/day of water_chems required
    water_chems_rate = (
        units.convert(tsa.flow_mol_in_total, to_units=units.kmol / units.hr)
        * units.hr
        / units.kmol
        * 0.1811
        / 1629629
        * units.ton
        / units.day
    )
    # ft^3/day of sorbent required
    sorbent_rate = (
        units.convert(tsa.bed_volume, to_units=units.ft**3)
        * (1 - tsa.bed_voidage)
        * tsa.number_beds
        / 0.5
        / 365
        / units.day
    )
    # kW/day of aux_power required
    aux_power_rate = total_auxiliary_load * 24 / units.day

    # kg/day of steam required
    steam_rate = units.convert(tsa.flow_mass_steam, to_units=units.kg / units.day)

    # variables for rates
    fs.costing.net_power = Var(fs.time, initialize=690, units=units.MW)
    fs.costing.water = Var(fs.time, initialize=1.0, units=units.gallon / units.day)
    fs.costing.water_chems = Var(fs.time, initialize=1.0, units=units.ton / units.day)
    fs.costing.sorbent = Var(fs.time, initialize=1.0, units=units.ft**3 / units.day)
    fs.costing.aux_power = Var(fs.time, initialize=1.0, units=units.kW / units.day)
    fs.costing.steam = Var(fs.time, initialize=1.0, units=units.kg / units.day)

    # constraints for rates
    @fs.costing.Constraint(fs.time, doc="Equation for cost of water")
    def water_eq(b, t):
        return b.water[t] == water_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of water_chems")
    def water_chems_eq(b, t):
        return b.water_chems[t] == water_chems_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of sorbent")
    def sorbent_eq(b, t):
        return b.sorbent[t] == sorbent_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of aux_power")
    def aux_power_eq(b, t):
        return b.aux_power[t] == aux_power_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of steam")
    def steam_eq(b, t):
        return b.steam[t] == steam_rate

    fs.costing.steam_eq.deactivate()
    fs.costing.steam.fix(0.0)

    fs.costing.net_power.fix()

    # resources to be costed
    resources = [
        "water",
        "water_treatment_chemicals",
        "sorbent",
        "aux_power",
        "waste_sorbent",
        "IP_steam",
    ]

    # vars for resource consumption rates
    rates = [
        fs.costing.water,
        fs.costing.water_chems,
        fs.costing.sorbent,
        fs.costing.aux_power,
        fs.costing.sorbent,
        fs.costing.steam,
    ]

    # resource prices
    prices = {
        "sorbent": 201 * units.USD_2018 / units.ft**3,
        "aux_power": 0.06 * units.USD_2018 / units.kW,
        "waste_sorbent": 0.86 * units.USD_2018 / units.ft**3,
        "IP_steam": 0.00733 * units.USD_2018 / units.kg,
    }

    # land cost for total overnight cost
    @fs.costing.Expression(doc="Land Cost [$MM]")
    def land_cost_exp(b):
        return (156000 * (tsa.number_beds / 120) ** (0.78)) * 1e-6  # scaled to Millions

    # build fixed and variables O&M cost
    fs.costing.build_process_costs(
        net_power=fs.costing.net_power,
        # arguments related to fixed OM costs
        total_plant_cost=True,
        labor_rate=38.50,
        labor_burden=30,
        operators_per_shift=8,
        tech=6,
        fixed_OM=True,
        # arguments related owners costs
        variable_OM=True,
        land_cost=fs.costing.land_cost_exp,
        resources=resources,
        rates=rates,
        prices=prices,
        fuel=None,
        tonne_CO2_capture=tsa.total_CO2_captured_year,
    )

    # electric boiler emissions
    @fs.Expression(doc="Electric Boiler Emissions [mol/s]")
    def emissions_electric_boiler(b):
        # eq. emissions for a US average electricity grid - kgCO2e/MWh
        equi_emissions = 425.021
        return total_auxiliary_load * 1e-3 * equi_emissions * 1000 / 3600 / 44.01

    @fs.Expression(doc="Electric Boiler Emissions, PV electricity grid [mol/s]")
    def emissions_electric_boiler_pv(b):
        # eq. emissions for solar PV electricity grid - kgCO2e/MWh
        equi_emissions = 48.472
        return total_auxiliary_load * 1e-3 * equi_emissions * 1000 / 3600 / 44.01

    # initialization
    for c in fs.costing._registered_unit_costing:
        for key in c.bare_erected_cost.keys():
            calculate_variable_from_constraint(
                c.bare_erected_cost[key],
                c.bare_erected_cost_eq[key],
            )
            calculate_variable_from_constraint(
                c.total_plant_cost[key],
                c.total_plant_cost_eq[key],
            )
    calculate_variable_from_constraint(fs.costing.water[0], fs.costing.water_eq[0])
    calculate_variable_from_constraint(
        fs.costing.water_chems[0], fs.costing.water_chems_eq[0]
    )
    calculate_variable_from_constraint(fs.costing.sorbent[0], fs.costing.sorbent_eq[0])
    calculate_variable_from_constraint(
        fs.costing.aux_power[0], fs.costing.aux_power_eq[0]
    )
    fs.costing.initialize_fixed_OM_costs()
    fs.costing.initialize_variable_OM_costs()


def _get_costing_retrofit_ngcc(tsa):

    # load custom costing parameters
    with open(
        os.path.join(directory, "costing_params_dac_retrofit_ngcc.json"), "r"
    ) as f:
        costing_params = json.load(f)

    # get flowsheet
    fs = tsa.flowsheet()

    # create costing block
    fs.costing = QGESSCosting()

    # EPAT accounts for: 1) Sorbent Handling,
    # 2) Sorbent Preparation & Feed,
    # 7) HRSG, Ductwork & Stack, 8) Steam Turbine & Accessories
    # 10) Ash/Spent Sorbent Handling System, 11) Accessory Electric Plant
    # 12) Instrumentation & Control, 15) DAC system

    # grouping accounts
    sorbent_makeup_accounts = [
        "1.5",
        "1.6",
        "1.7",
        "1.8",
        "1.9",
        "2.5",
        "2.6",
        "2.9",
        "10.6",
        "10.7",
        "10.9",
    ]
    gas_flow_to_dac_accounts = ["7.3"]
    steam_accounts = ["8.4"]
    total_auxiliary_load_accounts = [
        "11.2",
        "11.3",
        "11.4",
        "11.5",
        "11.6",
        "12.1",
        "12.2",
        "12.3",
        "12.4",
        "12.5",
        "12.6",
        "12.7",
        "12.8",
        "12.9",
    ]
    dac_system_accounts = [
        "15.1",
        "15.2",
        "15.3",
        "15.4",
        "15.5",
        "15.6",
        "15.7",
        "15.8",
    ]

    # reference parameters for accounts
    # sorbent makeup rate - from model
    sorbent_makeup_rate = (
        units.convert(tsa.bed_volume, to_units=units.ft**3)
        * (1 - tsa.bed_voidage)
        * tsa.number_beds
        / 0.5
        / 365
        / units.d
    )  # [ft^3/d]
    # CO2 product mass flow rate - from model
    CO2_product_mass_flow = units.convert(
        tsa.mw["CO2"] * tsa.flow_mol_co2_rich_stream[0, "CO2"],
        to_units=units.lb / units.hr,
    )  # [lb/hr]
    # mole flow rate of dac feed - from model (exhaust or air)
    flow_mol_inlet = units.convert(
        tsa.flow_mol_in_total, to_units=units.kmol / units.hr
    )  # [kmol/hr]
    # mass flow rate of dac feed - from model (exhaust or air)
    flow_mass_inlet = units.convert(
        tsa.flow_mass_in_total, to_units=units.lb / units.hr
    )  # [lb/hr]
    # compressor auxiliary load - from surrogates
    product_compressor_auxiliary_load = (
        0.0012 * flow_mol_inlet * units.hr / units.kmol - 2.2798
    ) * units.kW  # [kW]
    # total auxiliary load
    # calculate with fans work + compressor for CO2 pure
    total_auxiliary_load = product_compressor_auxiliary_load + units.convert(
        tsa.compressor.unit.work_mechanical[0], to_units=units.kW
    )
    # compressor aftercooler heat exchanger duty - from surrogates
    compressor_aftercooler_heat_duty = (
        (2e-6 * flow_mol_inlet * units.hr / units.kmol - 7e-8) * units.MBtu / units.hr
    )  # [MMBtu/hr]
    # auxiliary load for 2 beds (pressure changer from TSA model estimates
    # the power required to move air or exhaust gas to all beds)
    auxiliary_load_2_beds = (
        2
        * units.convert(tsa.compressor.unit.work_mechanical[0], to_units=units.kW)
        / tsa.number_beds
    )  # [kW]

    # sorbent makeup accounts
    fs.sorbent_makeup = UnitModelBlock()
    fs.sorbent_makeup.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": sorbent_makeup_accounts,
            "scaled_param": sorbent_makeup_rate,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # gas flow to dac accounts
    fs.gas_flow_to_dac = UnitModelBlock()
    fs.gas_flow_to_dac.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": gas_flow_to_dac_accounts,
            "scaled_param": flow_mass_inlet,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # steam accounts
    fs.steam = UnitModelBlock()
    fs.steam.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": steam_accounts,
            "scaled_param": units.convert(
                tsa.flow_mass_steam, to_units=units.lb / units.hr
            ),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # total plant auxiliary load accounts
    fs.total_auxiliary_load = UnitModelBlock()
    fs.total_auxiliary_load.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": total_auxiliary_load_accounts,
            "scaled_param": total_auxiliary_load,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # get EPAT costing accounts for 15) DAC system
    accounts_info = {}
    for k, v in costing_params["8"]["B"].items():
        if k in dac_system_accounts:
            accounts_info[k] = v["Account Name"].replace("DAC ", "")
    accounts = list(accounts_info.keys())

    # 15.1 - DAC adsorption/desorption vessels (cost of 120 vessels)
    vessels_account = [accounts[0]]
    fs.vessels = UnitModelBlock()
    fs.vessels.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": vessels_account,
            "scaled_param": units.convert(tsa.bed_volume, to_units=units.ft**3),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.2 - DAC CO2 Compression & Drying
    product_compression_account = [accounts[1]]
    fs.product_compression = UnitModelBlock()
    fs.product_compression.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": product_compression_account,
            "scaled_param": product_compressor_auxiliary_load,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.3 - DAC CO2 Compressor Aftercooler
    compressor_aftercooler_account = [accounts[2]]
    fs.compressor_aftercooler = UnitModelBlock()
    fs.compressor_aftercooler.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": compressor_aftercooler_account,
            "scaled_param": compressor_aftercooler_heat_duty,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.4 - DAC System Air Handling Duct and Dampers
    duct_dampers_account = [accounts[3]]
    fs.duct_dampers = UnitModelBlock()
    fs.duct_dampers.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": duct_dampers_account,
            "scaled_param": 2
            * units.convert(tsa.flow_mass_in_total_bed, to_units=units.lb / units.hr),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.5 - DAC System Air Handling Fans
    feed_fans_account = [accounts[4]]
    fs.feed_fans = UnitModelBlock()
    fs.feed_fans.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": feed_fans_account,
            "scaled_param": auxiliary_load_2_beds,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.6 - DAC Desorption Process Gas Handling System
    desorption_gas_handling_account = [accounts[5]]
    fs.desorption_gas_handling = UnitModelBlock()
    fs.desorption_gas_handling.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": desorption_gas_handling_account,
            "scaled_param": CO2_product_mass_flow,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.7 - DAC Steam Distribution System
    steam_distribution_account = [accounts[6]]
    fs.steam_distribution = UnitModelBlock()
    fs.steam_distribution.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": steam_distribution_account,
            "scaled_param": units.convert(
                tsa.flow_mass_steam, to_units=units.lb / units.hr
            ),
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # 15.8 - DAC System Controls Equipment
    controls_equipment_account = [accounts[7]]
    fs.controls_equipment = UnitModelBlock()
    fs.controls_equipment.costing = UnitModelCostingBlock(
        flowsheet_costing_block=fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": controls_equipment_account,
            "scaled_param": total_auxiliary_load,
            "tech": 8,
            "ccs": "B",
            "additional_costing_params": costing_params,
        },
    )

    # total plant cost
    TPC_list = {}
    for o in fs.component_objects(descend_into=True):
        # look for costing blocks
        if hasattr(o, "costing") and hasattr(o.costing, "total_plant_cost"):
            for k in o.costing.total_plant_cost.keys():
                if k not in ["15.1", "15.4", "15.5"]:
                    TPC_list[k] = o.costing.total_plant_cost[k]
                if k in ["15.1"]:
                    TPC_list[k] = o.costing.total_plant_cost[k] / 120 * tsa.number_beds
                if k in ["15.4", "15.5"]:
                    TPC_list[k] = o.costing.total_plant_cost[k] * tsa.number_beds / 2

    # Total plant cost of dac unit
    @fs.costing.Expression(doc="total TPC for TSA system in $MM")
    def total_TPC(b):
        return sum(TPC_list.values())

    # build variable costs components

    # gallon/day of water required
    water_rate = (
        units.convert(tsa.flow_mol_in_total, to_units=units.kmol / units.hr)
        * units.hr
        / units.kmol
        * 60792.407
        / 1629629
        * units.gallon
        / units.day
    )
    # ton/day of water_chems required
    water_chems_rate = (
        units.convert(tsa.flow_mol_in_total, to_units=units.kmol / units.hr)
        * units.hr
        / units.kmol
        * 0.1811
        / 1629629
        * units.ton
        / units.day
    )
    # ft^3/day of sorbent required
    sorbent_rate = (
        units.convert(tsa.bed_volume, to_units=units.ft**3)
        * (1 - tsa.bed_voidage)
        * tsa.number_beds
        / 0.5
        / 365
        / units.day
    )
    # kW/day of aux_power required
    aux_power_rate = total_auxiliary_load * 24 / units.day

    # kg/day of steam required
    steam_rate = units.convert(tsa.flow_mass_steam, to_units=units.kg / units.day)

    # variables for rates
    fs.costing.net_power = Var(fs.time, initialize=690, units=units.MW)
    fs.costing.water = Var(fs.time, initialize=1.0, units=units.gallon / units.day)
    fs.costing.water_chems = Var(fs.time, initialize=1.0, units=units.ton / units.day)
    fs.costing.sorbent = Var(fs.time, initialize=1.0, units=units.ft**3 / units.day)
    fs.costing.aux_power = Var(fs.time, initialize=1.0, units=units.kW / units.day)
    fs.costing.steam = Var(fs.time, initialize=1.0, units=units.kg / units.day)

    # constraints for rates
    @fs.costing.Constraint(fs.time, doc="Equation for cost of water")
    def water_eq(b, t):
        return b.water[t] == water_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of water_chems")
    def water_chems_eq(b, t):
        return b.water_chems[t] == water_chems_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of sorbent")
    def sorbent_eq(b, t):
        return b.sorbent[t] == sorbent_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of aux_power")
    def aux_power_eq(b, t):
        return b.aux_power[t] == aux_power_rate

    @fs.costing.Constraint(fs.time, doc="Equation for cost of steam")
    def steam_eq(b, t):
        return b.steam[t] == steam_rate

    fs.costing.steam_eq.deactivate()
    fs.costing.steam.fix(0.0)

    fs.costing.aux_power_eq.deactivate()
    fs.costing.aux_power.fix(0.0)

    fs.costing.net_power.fix()

    # resources to be costed
    resources = [
        "water",
        "water_treatment_chemicals",
        "sorbent",
        "aux_power",
        "waste_sorbent",
        "IP_steam",
    ]

    # vars for resource consumption rates
    rates = [
        fs.costing.water,
        fs.costing.water_chems,
        fs.costing.sorbent,
        fs.costing.aux_power,
        fs.costing.sorbent,
        fs.costing.steam,
    ]

    # resource prices
    prices = {
        "sorbent": 201 * units.USD_2018 / units.ft**3,
        "aux_power": 0.06 * units.USD_2018 / units.kW,
        "waste_sorbent": 0.86 * units.USD_2018 / units.ft**3,
        "IP_steam": 0.00733 * units.USD_2018 / units.kg,
    }

    # land cost for total overnight cost
    @fs.costing.Expression(doc="Land Cost [$MM]")
    def land_cost_exp(b):
        return (156000 * (tsa.number_beds / 120) ** (0.78)) * 1e-6  # scaled to Millions

    # build fixed and variables O&M cost
    fs.costing.build_process_costs(
        net_power=fs.costing.net_power,
        # arguments related to fixed OM costs
        total_plant_cost=True,
        labor_rate=38.50,
        labor_burden=30,
        operators_per_shift=8,
        tech=6,
        fixed_OM=True,
        # arguments related owners costs
        variable_OM=True,
        land_cost=fs.costing.land_cost_exp,
        resources=resources,
        rates=rates,
        prices=prices,
        fuel=None,
        tonne_CO2_capture=tsa.total_CO2_captured_year,
    )


def print_dac_costing(tsa):
    fs = tsa.flowsheet()

    TPC_list = {}
    for o in fs.component_objects(descend_into=True):
        # look for costing blocks
        if hasattr(o, "costing") and hasattr(o.costing, "total_plant_cost"):
            for k in o.costing.total_plant_cost.keys():
                if k not in ["15.1", "15.4", "15.5"]:
                    TPC_list[k] = o.costing.total_plant_cost[k]
                if k in ["15.1"]:
                    TPC_list[k] = o.costing.total_plant_cost[k] / 120 * tsa.number_beds
                if k in ["15.4", "15.5"]:
                    TPC_list[k] = o.costing.total_plant_cost[k] * tsa.number_beds / 2

    for i, k in TPC_list.items():
        print(i, value(k))


def _var_dict_costing(tsa):

    # get flowsheet
    fs = tsa.flowsheet()

    # create dir with costing summary
    var_dict = {}

    var_dict["Annualized capital cost of dac unit [$MM/year]"] = value(
        fs.costing.annualized_cost
    )
    var_dict["Fixed O&M cost of dac unit [$MM/year]"] = value(
        fs.costing.total_fixed_OM_cost
    )
    var_dict["Variable O&M cost of dac unit [$MM/year]"] = value(
        fs.costing.total_variable_OM_cost[0]
    )
    var_dict["Total annualized cost of dac unit [$MM/year]"] = value(
        fs.costing.annualized_cost
        + fs.costing.total_fixed_OM_cost
        + fs.costing.total_variable_OM_cost[0] * fs.costing.capacity_factor
    )
    var_dict["Capture cost [$/tonne CO2]"] = value(fs.costing.cost_of_capture * 1e6)

    if hasattr(fs, "emissions_electric_boiler"):
        var_dict["Electric Boiler Emissions [mol/s]"] = value(
            fs.emissions_electric_boiler
        )

    if hasattr(fs, "emissions_electric_boiler_pv"):
        var_dict["Electric Boiler Emissions, PV electricity grid [mol/s]"] = value(
            fs.emissions_electric_boiler_pv
        )

    return var_dict


def dac_costing_summary(tsa, export=False):

    fs = tsa.flowsheet()

    if not hasattr(fs, "vessels"):
        raise ConfigurationError(f"{tsa.name} does not have any costing block.")

    var_dict = _var_dict_costing(tsa)

    summary_dir = {}
    summary_dir["Value"] = {}
    summary_dir["pos"] = {}

    count = 1
    for k, v in var_dict.items():
        summary_dir["Value"][k] = value(v)
        summary_dir["pos"][k] = count
        count += 1

    df = DataFrame.from_dict(summary_dir, orient="columns")
    del df["pos"]
    if export:
        df.to_csv(f"{tsa.local_name}_summary_costing.csv")

    print("\n" + "=" * 84)
    print(f"summary costing {tsa.local_name}")
    print("-" * 84)
    stdout.write(textwrap.indent(stream_table_dataframe_to_string(df), " " * 4))
    print("\n" + "=" * 84 + "\n")
